# Generated by Django 5.0.6 on 2024-08-05 00:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("slurpit_netbox", "0012_slurpitprefix"),
    ]

    operations = [
        migrations.AddField(
            model_name="slurpitinitipaddress",
            name="ignore_status",
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 5.0.6 on 2024-08-05 01:39

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("slurpit_netbox", "0014_slurpitinitipaddress_ignore_description_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="slurpitinterface",
            name="ignore_duplex",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="slurpitinterface",
            name="ignore_module",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="slurpitinterface",
            name="ignore_speed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="slurpitinterface",
            name="ignore_type",
            field=models.BooleanField(default=False),
        ),
    ]



from core.choices import DataSourceStatusChoices
from django import forms
from django.utils.translation import gettext_lazy as _
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from tenancy.models import TenantGroup, Tenant
from tenancy.forms import TenancyForm
from .models import SlurpitImportedDevice, SlurpitPlanning, SlurpitSetting, SlurpitInitIPAddress, SlurpitInterface, SlurpitPrefix, SlurpitVLAN
from .management.choices import SlurpitApplianceTypeChoices
from extras.models import CustomField
from virtualization.models import VMInterface

from netbox.api.fields import ChoiceField
from netbox.forms import NetBoxModelBulkEditForm, NetBoxModelFilterSetForm, NetBoxModelForm

from ipam.models import FHRPGroup, VRF, IPAddress, VLANGroup, VLAN, Role
from ipam.choices import *
from ipam.constants import *
from ipam.forms import PrefixForm

from dcim.forms.common import InterfaceCommonForm
from dcim.choices import DeviceStatusChoices, DeviceAirflowChoices, DeviceStatusChoices, InterfaceSpeedChoices
from dcim.models import DeviceRole, DeviceType, Site, Location, Region, Rack, Device, Interface, Module, MACAddress

from utilities.forms import form_from_model, add_blank_choice
from utilities.forms.fields import CommentField, DynamicModelChoiceField, DynamicModelMultipleChoiceField
from utilities.forms.widgets import APISelect, NumberWithOptions, HTMXSelect

class DeviceComponentForm(NetBoxModelForm):
    device = DynamicModelChoiceField(
        label=_('Device'),
        queryset=Device.objects.all(),
        selector=True,
        required=False,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Disable reassignment of Device when editing an existing instance
        if self.instance.pk:
            self.fields['device'].disabled = False

class ModularDeviceComponentForm(DeviceComponentForm):
    module = DynamicModelChoiceField(
        label=_('Module'),
        queryset=Module.objects.all(),
        required=False,
        query_params={
            'device_id': '$device',
        }
    )



class OnboardingForm(NetBoxModelBulkEditForm):
    model = SlurpitImportedDevice
    interface_name = forms.CharField(
        label=_('Management Interface'),
        initial='Management1',
        max_length=200,
        required=True
    )
    status = forms.ChoiceField(
        label=_('Status'),
        choices=add_blank_choice(DeviceStatusChoices),
        required=False
    )
    device_type = forms.ChoiceField(
        choices=[],
        label=_('Device type'),
        required=True
    )
    role = DynamicModelChoiceField(
        label=_('Device role'),
        queryset=DeviceRole.objects.all(),
        required=True
    )
    site = forms.ChoiceField(
        choices=[],
        label=_('Site'),
        required=True
    )
    region = DynamicModelChoiceField(
        label=_('Region'),
        queryset=Region.objects.all(),
        required=False
    )
    location = DynamicModelChoiceField(
        label=_('Location'),
        queryset=Location.objects.all(),
        required=False,
        query_params={
            'site_id': '$site'
        },
        initial_params={
            'racks': '$rack'
        }
    )
    rack = DynamicModelChoiceField(
        label=_('Rack'),
        queryset=Rack.objects.all(),
        required=False,
        query_params={
            'site_id': '$site',
            'location_id': '$location',
        }
    )
    position = forms.DecimalField(
        label=_('Position'),
        required=False,
        help_text=_("The lowest-numbered unit occupied by the device"),
        localize=True,
        widget=APISelect(
            api_url='/api/dcim/racks/{{rack}}/elevation/',
            attrs={
                'disabled-indicator': 'device',
                'data-dynamic-params': '[{"fieldName":"face","queryParam":"face"}]'
            },
        )
    )
    latitude = forms.DecimalField(
        label=_('Latitude'),
        max_digits=8,
        decimal_places=6,
        required=False,
        help_text=_("GPS coordinate in decimal format (xx.yyyyyy)")
    )
    longitude = forms.DecimalField(
        label=_('longitude'),
        max_digits=9,
        decimal_places=6,
        required=False,
        help_text=_("GPS coordinate in decimal format (xx.yyyyyy)")
    )
    tenant_group = DynamicModelChoiceField(
        label=_('Tenant group'),
        queryset=TenantGroup.objects.all(),
        required=False,
        null_option='None',
        initial_params={
            'tenants': '$tenant'
        }
    )
    tenant = DynamicModelChoiceField(
        label=_('Tenant'),
        queryset=Tenant.objects.all(),
        required=False,
        query_params={
            'group_id': '$tenant_group'
        }
    )
    description = forms.CharField(
        label=_('Description'),
        max_length=200,
        required=False
    )
    airflow = forms.ChoiceField(
        label=_('Airflow'),
        choices=add_blank_choice(DeviceAirflowChoices),
        required=False
    )
    
    def __init__(self, *args, **kwargs):
        device_types = kwargs['initial'].pop('device_types', None)
        sites = kwargs['initial'].pop('sites', None)

        super().__init__(*args, **kwargs)

        choices = []
        if device_types and len(device_types) > 1:
            choices = [('keep_original', 'Keep Original Type')]
        for dt in DeviceType.objects.all().order_by('id'):
            choices.append((dt.id, dt.model))          
        self.fields['device_type'].choices = choices

        choices = []
        if sites and len(sites) > 1:
            choices = [('keep_original', 'Keep Original Type')]

        for site in Site.objects.all().order_by('id'):
            choices.append((site.id, site.name))          
        self.fields['site'].choices = choices

class SlurpitPlanningTableForm(forms.Form):
    planning_id = DynamicModelChoiceField(
        queryset=SlurpitPlanning.objects.all(),
        to_field_name='planning_id',
        required=True,
        label=_("Slurpit Plans"),
    )

class SlurpitApplianceTypeForm(forms.Form):
    model =  SlurpitSetting
    appliance_type = forms.ChoiceField(
        label=_('Data synchronization'),
        choices=add_blank_choice(SlurpitApplianceTypeChoices),
        required=False
    )

class SlurpitMappingForm(forms.Form):
    source_field = forms.CharField(
        required=True,
        label=_("Source Field"),
        widget=forms.TextInput(attrs={"class": "form-control"}),
        help_text=_("Slurpit Device's Field"),
    )
    
    target_field = forms.ChoiceField(
        choices=[
        ],
        required=True,
        label=f"Target Field",
    )
    
    
    def __init__(self, *args, **kwargs):
        choice_name = kwargs.pop('choice_name', None) 
        doaction = kwargs.pop('doaction', None) 
        super(SlurpitMappingForm, self).__init__(*args, **kwargs)
        
        choices = []
        
        for field in Device._meta.get_fields():
            if not field.is_relation or field.one_to_one or (field.many_to_one and field.related_model):
                choices.append((f'device|{field.name}', f'device | {field.name}'))
        
        # Add custom fields
        device = ContentType.objects.get(app_label='dcim', model='device')
        device_custom_fields = CustomField.objects.filter(object_types=device)

        for custom_field in device_custom_fields:
            choices.append((f'device|cf_{custom_field.name}', f'device | {custom_field.name}'))
        
        self.fields[f'target_field'].choices = choices

        if doaction != "add":
            self.fields[f'target_field'].label = choice_name
            del self.fields[f'source_field']


class SlurpitDeviceForm(forms.Form):
    device = DynamicModelChoiceField(
        queryset=Device.objects.all(),
        required=True,
        label=_("Device"),
    )

class SlurpitDeviceStatusForm(forms.Form):
    device_status = forms.ChoiceField(
        label=_('Status'),
        choices=add_blank_choice(DeviceStatusChoices),
        required=False
    )

class SlurpitInitIPAMForm(TenancyForm, NetBoxModelForm):
    vrf = DynamicModelChoiceField(
        queryset=VRF.objects.all(),
        required=False,
        label=_('VRF')
    )
    enable_reconcile = forms.BooleanField(
        required=False,
        label=_('Enable to reconcile every incoming IPAM data')
    )
    ignore_status = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_role = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data'),
    )
    ignore_vrf = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data'),
    )
    ignore_tenant = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data'),
    )
    ignore_description = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data'),
    )

    comments = CommentField()

    class Meta:
        model = SlurpitInitIPAddress
        fields = [
            'vrf', 'status', 'ignore_status', 'role', 'tenant_group',
            'tenant','description', 'enable_reconcile','ignore_role', 'ignore_tenant', 'ignore_description'
        ]

    def __init__(self, *args, **kwargs):
        # Initialize helper selectors
        initial = kwargs.get('initial', {}).copy()
        kwargs['initial'] = initial

        super().__init__(*args, **kwargs)
        del self.fields['tags']
    
class SlurpitInitIPAMEditForm(SlurpitInitIPAMForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        del self.fields['enable_reconcile']

class SlurpitDeviceInterfaceForm(InterfaceCommonForm, ModularDeviceComponentForm):
    enable_reconcile = forms.BooleanField(
        required=False,
        label=_('Enable to reconcile every incoming Device Interface data')
    )
    
    vlan_group = DynamicModelChoiceField(
        queryset=VLANGroup.objects.all(),
        required=False,
        label=_('VLAN group')
    )

    untagged_vlan = DynamicModelChoiceField(
        queryset=VLAN.objects.all(),
        required=False,
        label=_('Untagged VLAN'),
        query_params={
            'group_id': '$vlan_group',
            'available_on_device': '$device',
        }
    )

    tagged_vlans = DynamicModelMultipleChoiceField(
        queryset=VLAN.objects.all(),
        required=False,
        label=_('Tagged VLANs'),
        query_params={
            'group_id': '$vlan_group',
            'available_on_device': '$device',
        }
    )
    qinq_svlan = DynamicModelChoiceField(
        queryset=VLAN.objects.all(),
        required=False,
        label=_('Q-in-Q Service VLAN'),
        query_params={
            'group_id': '$vlan_group',
            'available_on_virtualmachine': '$virtual_machine',
            'qinq_role': VLANQinQRoleChoices.ROLE_SERVICE,
        }
    )
    primary_mac_address = DynamicModelChoiceField(
        queryset=MACAddress.objects.all(),
        label=_('Primary MAC address'),
        required=False,
        quick_add=True,
        quick_add_params={'interface': '$pk'}
    )

    ignore_module = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_type = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_speed = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_duplex = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )

    class Meta:
        model = SlurpitInterface
        fields = [
           'module', 'name', 'label', 'type', 'speed', 'duplex',  'description', 'mode', 'vlan_group', 'untagged_vlan', 'tagged_vlans', 'enable_reconcile',
            'ignore_module', 'ignore_type', 'ignore_speed', 'ignore_duplex'
        ]
        widgets = {
            'speed': NumberWithOptions(
                options=InterfaceSpeedChoices
            ),
            'mode': HTMXSelect(),
        }
        labels = {
            'mode': '802.1Q Mode',
        }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        del self.fields['tags']
        del self.fields['device']

class SlurpitDeviceInterfaceEditForm(SlurpitDeviceInterfaceForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        del self.fields['enable_reconcile']

class SlurpitPrefixForm(PrefixForm):
    enable_reconcile = forms.BooleanField(
        required=False,
        label=_('Enable to reconcile every incoming Prefix data')
    )
    ignore_status = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_vrf = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_role = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_site = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_vlan = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_tenant = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_description = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    
    class Meta:
        model = SlurpitPrefix
        fields = [
            'prefix', 'vrf', 'scope_type', '_site', '_site_group', '_location', '_region', 'vlan', 'status', 'role', 'is_pool', 'mark_utilized', 'tenant_group', 'tenant',
            'description', 'comments', 'tags','enable_reconcile', 'ignore_status', 'ignore_vrf', 'ignore_role', 'ignore_site', 'ignore_vlan',
            'ignore_tenant', 'ignore_description'
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        del self.fields['tags']
class ComponentBulkEditForm(NetBoxModelBulkEditForm):
    device = forms.ModelChoiceField(
        label=_('Device'),
        queryset=Device.objects.all(),
        required=False,
        disabled=True,
        widget=forms.HiddenInput()
    )
    module = forms.ModelChoiceField(
        label=_('Module'),
        queryset=Module.objects.all(),
        required=False
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Limit module queryset to Modules which belong to the parent Device
        if 'device' in self.initial:
            device = Device.objects.filter(pk=self.initial['device']).first()
            self.fields['module'].queryset = Module.objects.filter(device=device)
        else:
            self.fields['module'].choices = ()
            self.fields['module'].widget.attrs['disabled'] = True

class SlurpitInterfaceBulkEditForm(
    form_from_model(SlurpitInterface, [
        'label',  'type', 'speed', 'duplex',  'description', 'mode'
    ]),
    ComponentBulkEditForm
):
    # enable_reconcile = forms.BooleanField(
    #     required=False,
    #     label=_('Enable to reconcile every incoming Device Interface data')
    # )
    
    model = SlurpitInterface

    nullable_fields = (
        'module', 'label', 'parent', 'bridge', 'lag', 'speed', 'duplex', 'mac_address', 'wwn', 'vdcs', 'mtu',
        'description', 'poe_mode', 'poe_type', 'mode', 'rf_channel', 'rf_channel_frequency', 'rf_channel_width',
        'tx_power', 'untagged_vlan', 'tagged_vlans', 'vrf', 'wireless_lans'
    )

    fields = [
        'module', 'name', 'label', 'type', 'speed', 'duplex',  'description', 'mode',
    ]


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        del self.fields['add_tags']
        del self.fields['remove_tags']

class SlurpitPrefixBulkEditForm(
    form_from_model(SlurpitPrefix, [
        'vrf', 'status', 'role','tenant', 'description', 'comments'
    ]),
    NetBoxModelBulkEditForm
):
    # enable_reconcile = forms.BooleanField(
    #     required=False,
    #     label=_('Enable to reconcile every incoming Prefix data')
    # )

    nullable_fields = (
        'vrf', 'tenant', 'role', 'description', 'comments',
    )
    tenant = DynamicModelChoiceField(
        label=_('Tenant'),
        queryset=Tenant.objects.all(),
        required=False
    )
    role = DynamicModelChoiceField(
        label=_('Role'),
        queryset=Role.objects.all(),
        required=False
    )
    vrf = DynamicModelChoiceField(
        queryset=VRF.objects.all(),
        required=False,
        label=_('VRF')
    )
    model = SlurpitPrefix
    fields = [
       'vrf', 'vlan', 'status', 'role', 'is_pool', 'mark_utilized', 'tenant_group', 'tenant',
        'description', 'comments', 'tags'
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        del self.fields['add_tags']
        del self.fields['remove_tags']
class SlurpitIPAddressBulkEditForm(
    form_from_model(SlurpitInitIPAddress, [
        'vrf', 'status', 'role', 'tenant', 'description'
    ]),
    NetBoxModelBulkEditForm
):
    tenant = DynamicModelChoiceField(
        label=_('Tenant'),
        queryset=Tenant.objects.all(),
        required=False
    )
    vrf = DynamicModelChoiceField(
        queryset=VRF.objects.all(),
        required=False,
        label=_('VRF')
    )
    nullable_fields = (
        'vrf', 'role', 'tenant', 'dns_name', 'description', 'comments',
    )
    model = SlurpitInitIPAddress
    fields = [
        'vrf', 'status', 'role', 'tenant_group',
        'tenant','description'
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        del self.fields['add_tags']
        del self.fields['remove_tags']

class SlurpitVLANForm(TenancyForm, NetBoxModelForm):
    site = DynamicModelChoiceField(
        label=_('Site'),
        queryset=Site.objects.all(),
        required=False,
        null_option='None',
        selector=True
    )
    role = DynamicModelChoiceField(
        label=_('Role'),
        queryset=Role.objects.all(),
        required=False
    )

    enable_reconcile = forms.BooleanField(
        required=False,
        label=_('Enable to reconcile every incoming Prefix data')
    )

    ignore_status = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_site = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_group = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_vid = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_role = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_tenant = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    ignore_description = forms.BooleanField(
        required=False,
        label=_('Ignore value for updating data')
    )
    class Meta:
        model = SlurpitVLAN
        fields = [
            'enable_reconcile', 
            'site', 
            'group', 
            'vid', 
            'name', 
            'status', 
            'role', 
            'tenant_group', 
            'tenant', 
            'description',
            'ignore_status',
            'ignore_site',
            'ignore_group',
            'ignore_vid',
            'ignore_role',
            'ignore_tenant',
            'ignore_description'
        ]

class SlurpitVLANEditForm(SlurpitVLANForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        del self.fields['enable_reconcile']

class SlurpitVLANBulkEditForm(
    form_from_model(SlurpitVLAN, [
        'status', 'role', 'tenant', 'description'
    ]),
    NetBoxModelBulkEditForm
):
    tenant = DynamicModelChoiceField(
        label=_('Tenant'),
        queryset=Tenant.objects.all(),
        required=False
    )
    role = DynamicModelChoiceField(
        label=_('Role'),
        queryset=Role.objects.all(),
        required=False
    )
    nullable_fields = (
        'role', 'tenant','description',
    )
    model = SlurpitVLAN
    fields = [
        'status', 'role', 'role', 'tenant_group',
        'tenant','description'
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        del self.fields['add_tags']
        del self.fields['remove_tags']


# Generated by Django 4.2.9 on 2024-04-18 07:40

from django.db import migrations, models
import django.db.models.deletion
import taggit.managers
import utilities.json


class Migration(migrations.Migration):

    dependencies = [
        ('extras', '0105_customfield_min_max_values'),
        ('tenancy', '0014_contactassignment_ordering'),
        ('ipam', '0069_gfk_indexes'),
        ('slurpit_netbox', '0006_slurpitsnapshot_result_type'),
    ]

    operations = [
        migrations.CreateModel(
            name='SlurpitInitIPAddress',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('comments', models.TextField(blank=True)),
                ('status', models.CharField(default='active', max_length=50)),
                ('role', models.CharField(blank=True, max_length=50)),
                ('enable_reconcile', models.BooleanField(default=False)),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
                ('tenant', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='slurpit_ip_addresses', to='tenancy.tenant')),
                ('vrf', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='slurpit_ip_addresses', to='ipam.vrf')),
            ],
            options={
                'verbose_name': 'Slurpit Init IP address',
                'verbose_name_plural': 'Slurpit Init IP addresses',
            },
        ),
    ]



import django_filters
from core.choices import DataSourceStatusChoices
from django.db.models import Q
from netbox.filtersets import NetBoxModelFilterSet, BaseFilterSet
from .models import SlurpitPlanning, SlurpitSnapshot, SlurpitImportedDevice, SlurpitInitIPAddress, SlurpitInterface, SlurpitPrefix, SlurpitVLAN
from django.utils.translation import gettext as _
from utilities.filters import (
    ContentTypeFilter, MultiValueCharFilter
)
import netaddr
from netaddr.core import AddrFormatError

class SlurpitPlanningFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )

    class Meta:
        model = SlurpitPlanning
        fields = ["id", "name", "planning_id"]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value)
        )
    
class SlurpitSnapshotFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )

    class Meta:
        model = SlurpitSnapshot
        fields = ["id", "hostname", "planning_id"]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value)
        )

class SlurpitImportedDeviceFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )

    class Meta:
        model = SlurpitImportedDevice
        fields = ["id", "hostname", "device_os", "device_type", "fqdn", "brand", "ipv4", "site", "serial", "os_version", "snmp_uptime"]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(device_os__icontains=value) | 
            Q(hostname__icontains=value) | 
            Q(device_type__icontains=value) | 
            Q(fqdn__icontains=value) | 
            Q(brand__icontains=value) | 
            Q(ipv4__icontains=value) | 
            Q(serial__icontains=value) | 
            Q(os_version__icontains=value)
        )

class SlurpitPrefixFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )

    prefix = MultiValueCharFilter(
        method='filter_prefix',
        label=_('Prefix'),
    )

    class Meta:
        model = SlurpitPrefix
        fields = ["id", "description", "prefix", "vrf"]

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        qs_filter = Q(description__icontains=value)
        qs_filter |= Q(prefix__contains=value.strip())
        qs_filter |= Q(vrf__name__contains=value)

        if value in 'Global':
            qs_filter |= Q(vrf=None)
        try:
            prefix = str(netaddr.IPNetwork(value.strip()).cidr)
            qs_filter |= Q(prefix__net_contains_or_equals=prefix)
            qs_filter |= Q(prefix__contains=value.strip())
        except (AddrFormatError, ValueError):
            pass
        return queryset.filter(qs_filter)


class SlurpitIPAddressFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )

    address = MultiValueCharFilter(
        method='filter_address',
        label=_('Address'),
    )

    class Meta:
        model = SlurpitInitIPAddress
        fields = [
            'address', 'dns_name', 'description'
        ] 

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        
        return queryset.filter(
            Q(dns_name__icontains=value) | 
            Q(description__icontains=value) | 
            Q(address__icontains=value)
        )
    
class SlurpitInterfaceFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )

    class Meta:
        model = SlurpitInterface
        fields = [
            'name', 'device', 'type', 'description'
        ] 

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        
        return queryset.filter(
            Q(name__icontains=value) | 
            Q(description__icontains=value) | 
            Q(type__icontains=value) | 
            Q(device__name__icontains=value)
        )
    
class SlurpitVLANFilterSet(BaseFilterSet):
    q = django_filters.CharFilter(
        method='search',
        label=_('Search'),
    )

    class Meta:
        model = SlurpitVLAN
        fields = [
            'name', 'description'
        ] 

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        
        return queryset.filter(
            Q(name__icontains=value) | 
            Q(description__icontains=value)
        )


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slurpit_netbox', '0025_alter_slurpitprefix_site_rename'),
    ]

    operations = [
        migrations.AddField(
            model_name='slurpitimporteddevice',
            name='os_version',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='slurpitimporteddevice',
            name='serial',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='slurpitimporteddevice',
            name='snmp_uptime',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='slurpitstageddevice',
            name='os_version',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='slurpitstageddevice',
            name='serial',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='slurpitstageddevice',
            name='snmp_uptime',
            field=models.CharField(max_length=255, null=True),
        ),
    ]



from django.views.generic import View
from ..models import (
    SlurpitImportedDevice, 
    SlurpitMapping, 
    SlurpitSetting, 
    SlurpitInitIPAddress, 
    SlurpitInterface, 
    SlurpitPrefix,
    SlurpitVLAN
)
from .. import forms, importer, models, tables
from ..decorators import slurpit_plugin_registered
from django.utils.decorators import method_decorator
from django.shortcuts import render, redirect
from ..forms import SlurpitMappingForm, SlurpitDeviceForm, SlurpitDeviceStatusForm, SlurpitInitIPAMForm, SlurpitDeviceInterfaceForm, SlurpitPrefixForm, SlurpitVLANForm
from ..management.choices import *
from django.contrib import messages
from dcim.models import Device
from django.forms.models import model_to_dict
import requests
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse, JsonResponse
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.contrib.contenttypes.models import ContentType
from extras.models import CustomField
from extras.models.tags import Tag
from ipam.models import IPRange
from urllib.parse import urlencode
from utilities.forms import restrict_form_fields
from utilities.exceptions import AbortRequest, PermissionsViolation
from django.db import router, transaction

BATCH_SIZE = 128

def get_device_dict(instance):
    device_dict = model_to_dict(instance)
    # Assuming 'device_type' is a ForeignKey, for example.
    device_dict['device_type'] = str(instance.device_type) if instance.device_type is not None else None
    device_dict['platform'] = str(instance.platform) if instance.platform is not None else None
    device_dict['primary_ip4'] = str(instance.primary_ip4) if instance.primary_ip4 is not None else None
    device_dict['primary_ip6'] = str(instance.primary_ip6) if instance.primary_ip6 is not None else None

    for custom_field in device_dict['custom_field_data']:
        device_dict[f'cf_{custom_field}'] = device_dict['custom_field_data'][custom_field]

    return device_dict

def post_slurpit_device(row, device_name):
    try:
        setting = SlurpitSetting.objects.get()
        uri_base = setting.server_url
        headers = {
                        'Authorization': f'Bearer {setting.api_key}',
                        'useragent': 'netbox/requests',
                        'accept': 'application/json',
                        'Content-Type': 'application/json',
                    }

        uri_devices = f"{uri_base}/api/devices/sync"
        
        try:
            row["ignore_plugin"] = str(1)
            r = requests.post(uri_devices, headers=headers, json=row, timeout=15, verify=False)
            r = r.json()
            r["device_name"] = device_name
            return r
        except Exception as e:
            return {"error": str(e), "device_name": device_name}

    except ObjectDoesNotExist:
        setting = None
        
        return {"error": "Need to set the setting parameter", "device_name": device_name}
    
    return None

@method_decorator(slurpit_plugin_registered, name='dispatch')
class DataMappingView(View):
    template_name = "slurpit_netbox/data_mapping.html"
    app_label = "dcim"
    model_name = "device"

    def get(self, request):
        sync = request.GET.get('sync', None)
        tab = request.GET.get('tab', None)
        subtab = request.GET.get('subtab', None)
        if sync is not None:
            # request_body = []

            netbox_devices = Device.objects.all()
            devices_array = [get_device_dict(device) for device in netbox_devices]

            objs = SlurpitMapping.objects.all()
            
            for device in devices_array:
                row = {}
                for obj in objs:
                    target_field = obj.target_field.split('|')[1]
                    row[obj.source_field] = str(device[target_field]) if device[target_field] is not None else None

                    if obj.source_field == 'ipv4' or obj.source_field == 'fqdn':
                        row[obj.source_field] = row[obj.source_field].split('/')[0]
                # request_body.append(row)

                res = post_slurpit_device(row, device["name"])

                if res is None:
                    return redirect(f'{request.path}?tab={tab}')
                
                if res['status'] != 200:
                    for error in res["messages"]:
                        messages.error(request, f'{escape(error)}: {escape(res["messages"][error])}')

                    return redirect(f'{request.path}?tab={tab}')
                
            messages.success(request, "Sync from Netbox to Slurpit is done successfully.")
            return redirect(f'{request.path}?tab={tab}')


        form = [
        ]

        mappings = SlurpitMapping.objects.all()
        
        appliance_type = ''
        try:
            setting = SlurpitSetting.objects.get()
            appliance_type = setting.appliance_type
        except ObjectDoesNotExist:
            setting = None

        if tab == "devices":
            slurpit_tag = Tag.objects.get(slug="slurpit")
            ip_ranges = IPRange.objects.filter(tags__in=[slurpit_tag])
        
        for mapping in mappings:
            form.append({
                "choice": mapping,
                "form": SlurpitMappingForm(choice_name=mapping, initial={"target_field": mapping.target_field})
            })

        new_form = SlurpitMappingForm(doaction="add")
        device_form = SlurpitDeviceForm()
        device_status_form = SlurpitDeviceStatusForm()


        if tab == "slurpit_to_netbox":
            if subtab == None or subtab == 'ipam':
                obj = SlurpitInitIPAddress.objects.filter(address=None).first()
                if obj is not None:
                    form = SlurpitInitIPAMForm(instance=obj)
                else:
                    form = SlurpitInitIPAMForm(data={'enable_reconcile':True})
            elif subtab == 'prefix':
                obj = SlurpitPrefix.objects.filter(prefix=None).first()
                if obj is not None:
                    form = SlurpitPrefixForm(instance=obj)
                else:
                    form = SlurpitPrefixForm(data={'enable_reconcile':True, 'status': 'active'})
            elif subtab == 'vlan':
                obj = SlurpitVLAN.objects.filter(name='').first()
                if obj is not None:
                    form = SlurpitVLANForm(instance=obj)
                else:
                    form = SlurpitVLANForm(data={'enable_reconcile':True, 'status': 'active'})
            else:
                obj = SlurpitInterface.objects.filter(name='').first()

                if obj is not None:
                    form = SlurpitDeviceInterfaceForm(instance=obj)
                else:
                    form = SlurpitDeviceInterfaceForm(data={'type': 'other', 'enable_reconcile':True})

        return render(
            request,
            self.template_name, 
            {
                "form": form,
                "new_form": new_form,
                "device_form": device_form,
                "device_status_form": device_status_form,
                "appliance_type": appliance_type,
            }
        )
    
    def post(self, request):
        tab = request.GET.get('tab', None)
        mapping_type = ''

        if tab == "netbox_to_slurpit" or tab is None:
            tab = "netbox_to_slurpit"
            test = request.POST.get('test')
            device_id = request.POST.get('device_id')

            if device_id is not None:
                if device_id == "":
                    return JsonResponse({})
                
                device = Device.objects.get(id=int(device_id))
                device = get_device_dict(device)

                row = {}
                objs = SlurpitMapping.objects.all()
                for obj in objs:
                    target_field = obj.target_field.split('|')[1]
                    row[obj.source_field] = str(device[target_field]) if device[target_field] is not None else None
                    
                    if row[obj.source_field] is not None:
                        if obj.source_field == 'ipv4' or obj.source_field == 'fqdn' :
                            row[obj.source_field] = row[obj.source_field].split('/')[0]
                    
                if test is not None:
                    res = post_slurpit_device(row, device["name"])

                    if res is None:
                        return JsonResponse({"error": "Server Internal Error."})
                    
                    return JsonResponse(res)

                return JsonResponse(row)
            
            action = request.POST.get("action")
            if action is None:
                source_field = request.POST.get("source_field")
                target_field = request.POST.get("target_field")
                
                try:
                    obj= SlurpitMapping.objects.create(source_field=source_field, target_field=target_field)
                    log_message =f'Added a mapping  which {source_field} field converts to {target_field} field.'      
                    messages.success(request, log_message)
                except Exception as e:
                    log_message =f'Failted to add a mapping which {source_field} field converts to {target_field} field.'      
                    messages.error(request, log_message)
                    pass
                
                return redirect(f'{request.path}?tab={tab}')
            
            elif action == "delete":
                source_field_keys = request.POST.getlist('pk')
                SlurpitMapping.objects.filter(source_field__in=source_field_keys).delete()
                return redirect(f'{request.path}?tab={tab}')
            
            elif action == "save":
                source_fields = request.POST.getlist('source_field')
                target_fields = request.POST.getlist('target_field')
                count = len(source_fields)
                offset = 0
                qs = []
                for i in range(count):
                    mapping, created = SlurpitMapping.objects.get_or_create(
                        source_field=source_fields[i], 
                        defaults={'target_field': target_fields[i]}
                    )
                    if not created:
                        # If the object was retrieved and not created, you will update its `target_field`
                        mapping.target_field = target_fields[i]
                    qs.append(mapping)

                while offset < count:
                    batch_qs = qs[offset:offset + BATCH_SIZE]
                    to_import = []        
                    for maping in batch_qs:
                        to_import.append(maping)

                    SlurpitMapping.objects.bulk_update(to_import, fields={'target_field'})
                    offset += BATCH_SIZE
                    
                return redirect(f'{request.path}?tab={tab}')
            elif action == "sync":
                device_status = request.POST.get('status')
                if device_status == "":
                    netbox_devices = Device.objects.all().values("id", "status")
                else:
                    netbox_devices = Device.objects.filter(status=device_status).values("id")
                device_names = []
                if netbox_devices:
                    for device in netbox_devices:
                        device_names.append(device['id'])
                
                return JsonResponse({"device": device_names})

        elif tab == "slurpit_to_netbox":
            mapping_type = request.POST.get('mappingtype')
            if mapping_type == 'ipam':
                obj = SlurpitInitIPAddress.objects.filter(address=None).first()
                if obj is None:
                    obj = SlurpitInitIPAddress()

                form = SlurpitInitIPAMForm(data=request.POST, instance=obj)
                restrict_form_fields(form, request.user)

                if form.is_valid():
                    try:
                        with transaction.atomic():
                            obj = form.save()
                            messages.success(request, "Updated the Slurpit IP Address Default values successfully.")
                    except (AbortRequest, PermissionsViolation) as e:
                        # logger.debug(e.message)
                        form.add_error(None, e.message)
                else:
                    messages.error(request, "Slurpit IP Address Form Validation Failed.")
                    pass
            elif mapping_type == 'interface':
                obj = SlurpitInterface.objects.filter(name='').first()
                if obj is None:
                    obj = SlurpitInterface()

                form = SlurpitDeviceInterfaceForm(data=request.POST, instance=obj)
                restrict_form_fields(form, request.user)

                if form.is_valid():
                    try:
                        with transaction.atomic():
                            obj = form.save()
                            messages.success(request, "Updated the Slurpit Interface Default values successfully.")
                    except (AbortRequest, PermissionsViolation) as e:
                        # logger.debug(e.message)
                        form.add_error(None, e.message)
                else:
                    messages.error(request, "Slurpit Interface Form Validation Failed.")
                    pass
            
            elif mapping_type == 'prefix':
                obj = SlurpitPrefix.objects.filter(prefix=None).first()
                if obj is None:
                    obj = SlurpitPrefix()

                form = SlurpitPrefixForm(data=request.POST, instance=obj)
                restrict_form_fields(form, request.user)

                if form.is_valid():
                    try:
                        with transaction.atomic():
                            obj = form.save()
                            messages.success(request, "Updated the Slurpit Prefix Default values successfully.")
                    except (AbortRequest, PermissionsViolation) as e:
                        # logger.debug(e.message)
                        form.add_error(None, e.message)
                else:
                    messages.error(request, "Slurpit Prefix Form Validation Failed.")
                    pass
            elif mapping_type == 'vlan':
                obj = SlurpitVLAN.objects.filter(name='').first()
                if obj is None:
                    obj = SlurpitVLAN()
                form = SlurpitVLANForm(data=request.POST, instance=obj)
                restrict_form_fields(form, request.user)

                if form.is_valid():
                    try:
                        with transaction.atomic():
                            obj = form.save()
                            messages.success(request, "Updated the Slurpit VLAN Default values successfully.")
                    except (AbortRequest, PermissionsViolation) as e:
                        # logger.debug(e.message)
                        form.add_error(None, e.message)
                else:
                    form_errors = form.errors
                    print(form_errors)
                    messages.error(request, "Slurpit VLAN Form Validation Failed.")

        base_url = request.path

        if mapping_type == "":
            query_string = urlencode({'tab': tab})
        else:
            query_string = urlencode({'tab': tab, 'subtab': mapping_type})
        url = f'{base_url}?{query_string}'
        return redirect(url)


import netaddr
from netbox.models import PrimaryModel
from django.db import models
from ipam.fields import IPAddressField
from ipam.choices import IPAddressStatusChoices, IPAddressRoleChoices
from ipam.constants import IPADDRESS_ASSIGNMENT_MODELS, IPADDRESS_ROLES_NONUNIQUE
from django.contrib.contenttypes.fields import GenericForeignKey
from ipam.validators import DNSValidator
from ipam.managers import IPAddressManager
from django.db.models.functions import Cast
from ipam.lookups import Host
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.core.exceptions import ValidationError
from netbox.config import get_config
from urllib.parse import urlencode

class SlurpitInitIPAddress(PrimaryModel):
    """
    An IPAddress represents an individual IPv4 or IPv6 address and its mask. The mask length should match what is
    configured in the real world. (Typically, only loopback interfaces are configured with /32 or /128 masks.) Like
    Prefixes, IPAddresses can optionally be assigned to a VRF. An IPAddress can optionally be assigned to an Interface.
    Interfaces can have zero or more IPAddresses assigned to them.

    An IPAddress can also optionally point to a NAT inside IP, designating itself as a NAT outside IP. This is useful,
    for example, when mapping public addresses to private addresses. When an Interface has been assigned an IPAddress
    which has a NAT outside IP, that Interface's Device can use either the inside or outside IP as its primary IP.
    """
    address = IPAddressField(
        verbose_name=_('address'),
        help_text=_('IPv4 or IPv6 address (with mask)'),
        blank=True,
        null=True
    )
    status = models.CharField(
        verbose_name=_('status'),
        max_length=50,
        choices=IPAddressStatusChoices,
        default=IPAddressStatusChoices.STATUS_ACTIVE,
        help_text=_('The operational status of this IP')
    )
    role = models.CharField(
        verbose_name=_('role'),
        max_length=50,
        choices=IPAddressRoleChoices,
        blank=True,
        help_text=_('The functional role of this IP')
    )

    vrf = models.ForeignKey(
        to='ipam.VRF',
        on_delete=models.PROTECT,
        related_name='slurpit_ip_addresses',
        blank=True,
        null=True,
        verbose_name=_('VRF')
    )
    tenant = models.ForeignKey(
        to='tenancy.Tenant',
        on_delete=models.PROTECT,
        related_name='slurpit_ip_addresses',
        blank=True,
        null=True
    )

    enable_reconcile = models.BooleanField(
        default=False,
        verbose_name=_('enable reconcile'),
    )

    dns_name = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        validators=[DNSValidator],
        verbose_name=_('DNS name'),
        help_text=_('Hostname or FQDN (not case-sensitive)')
    )

    ignore_status = models.BooleanField(
        default=False,
        null=True,
        verbose_name=_('ignore status'),
    )
    ignore_role = models.BooleanField(
        default=False,
        null=True,
        verbose_name=_('ignore role'),
    )
    ignore_vrf = models.BooleanField(
        default=False,
        null=True,
        verbose_name=_('ignore vrf'),
    )
    ignore_tenant = models.BooleanField(
        default=False,
        null=True,
        verbose_name=_('ignore tenant'),
    )
    ignore_description = models.BooleanField(
        default=False,
        null=True,
        verbose_name=_('ignore description'),
    )

    class Meta:
        verbose_name = _('Slurpit IP address')
        verbose_name_plural = _('Slurpit  IP addresses')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        return f"{self.address}"
    
    def clean(self):
        super().clean()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    def get_status_color(self):
        return IPAddressStatusChoices.colors.get(self.status)
    
    def get_absolute_url(self):
        return reverse('plugins:slurpit_netbox:reconcile_detail', args=[self.pk, 'ipam'])
    
    def get_edit_url(self):
        query_params = {'tab': "ipam"}
        base_url = reverse("plugins:slurpit_netbox:reconcile_list")
        # Encode your query parameters and append them to the base URL
        url_with_querystring = f"{base_url}?{urlencode(query_params)}"

        base_url = reverse('plugins:slurpit_netbox:slurpitipaddress_edit', args=[self.pk])
        query_params = {'return_url': url_with_querystring}
        url_with_querystring = f"{base_url}?{urlencode(query_params)}"

        return url_with_querystring


# Generated by Django 4.2.8 on 2024-01-25 15:51

from django.db import migrations, models
import django.db.models.deletion
import taggit.managers
import utilities.json


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('extras', '0105_customfield_min_max_values'),
        ('dcim', '0185_gfk_indexes'),
    ]

    operations = [
        migrations.CreateModel(
            name='SlurpitLog',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('log_time', models.DateTimeField(auto_now=True)),
                ('level', models.CharField(default='default', editable=False, max_length=100)),
                ('category', models.CharField(default='init', editable=False, max_length=50)),
                ('message', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='SlurpitStagedDevice',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('slurpit_id', models.BigIntegerField(unique=True)),
                ('disabled', models.BooleanField(default=False)),
                ('hostname', models.CharField(max_length=255, unique=True)),
                ('fqdn', models.CharField(max_length=128)),
                ('ipv4', models.CharField(max_length=23)),
                ('device_os', models.CharField(max_length=128)),
                ('device_type', models.CharField(max_length=255)),
                ('brand', models.CharField(max_length=255)),
                ('createddate', models.DateTimeField()),
                ('changeddate', models.DateTimeField()),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SlurpitSnapshot',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('comments', models.TextField(blank=True)),
                ('hostname', models.CharField(max_length=255)),
                ('planning_id', models.BigIntegerField()),
                ('content', models.JSONField()),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SlurpitSetting',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('comments', models.TextField(blank=True)),
                ('server_url', models.CharField(max_length=200)),
                ('api_key', models.CharField(editable=False, max_length=50)),
                ('last_synced', models.DateTimeField(auto_now=True, null=True)),
                ('connection_status', models.CharField(default='', editable=False, max_length=50, null=True)),
                ('push_api_key', models.CharField(editable=False, max_length=200, null=True)),
                ('appliance_type', models.CharField(blank=True, max_length=50)),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'setting',
                'verbose_name_plural': 'setting',
            },
        ),
        migrations.CreateModel(
            name='SlurpitPlanning',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('comments', models.TextField(blank=True)),
                ('name', models.CharField(max_length=255)),
                ('planning_id', models.BigIntegerField()),
                ('selected', models.BooleanField(default=False)),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SlurpitImportedDevice',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('slurpit_id', models.BigIntegerField(unique=True)),
                ('disabled', models.BooleanField(default=False)),
                ('hostname', models.CharField(max_length=255, unique=True)),
                ('fqdn', models.CharField(max_length=128)),
                ('ipv4', models.CharField(max_length=23)),
                ('device_os', models.CharField(max_length=128)),
                ('device_type', models.CharField(max_length=255)),
                ('brand', models.CharField(max_length=255)),
                ('createddate', models.DateTimeField()),
                ('changeddate', models.DateTimeField()),
                ('mapped_device', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='dcim.device')),
                ('mapped_devicetype', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='dcim.devicetype')),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import json
import traceback
from datetime import timedelta

from rest_framework.routers import APIRootView
from rest_framework_bulk import BulkCreateModelMixin, BulkDestroyModelMixin
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from rest_framework.views import APIView
from rest_framework import status, mixins

from django.db import transaction
from django.http import JsonResponse
from django.contrib.contenttypes.models import ContentType
from django.forms.models import model_to_dict
from django.utils import timezone
from django.db.models import Q, Count
from django.core.serializers import serialize

from .serializers import (
    SlurpitPlanningSerializer, 
    SlurpitSnapshotSerializer, 
    SlurpitImportedDeviceSerializer, 
    SlurpitPrefixSerializer, 
    SlurpitInterfaceSerializer, 
    SlurpitInitIPAddressSerializer,
    SlurpitVLANSerializer
)
from ..validator import (
    device_validator, 
    ipam_validator, 
    interface_validator, 
    prefix_validator,
    vlan_validator
)
from ..importer import process_import, import_devices, import_plannings, start_device_import, create_sites, sync_sites, BATCH_SIZE
from ..management.choices import *
from ..views.datamapping import get_device_dict
from ..references import base_name 
from ..references.generic import status_offline, SlurpitViewSet, status_decommissioning
from ..references.imports import * 
from ..models import (
    SlurpitPlanning, 
    SlurpitSnapshot, 
    SlurpitImportedDevice, 
    SlurpitStagedDevice,
    SlurpitMapping, 
    SlurpitInitIPAddress, 
    SlurpitInterface, 
    SlurpitPrefix,
    SlurpitVLAN
)
from ..filtersets import SlurpitPlanningFilterSet, SlurpitSnapshotFilterSet, SlurpitImportedDeviceFilterSet
from ..views.setting import sync_snapshot
from ipam.models import (
    FHRPGroup, VRF, IPAddress, VLAN, Role, Prefix, VLANGroup
)
from ipam.api.serializers import PrefixSerializer
from dcim.models import Interface, Site
from dcim.forms import InterfaceForm
from ipam.forms import (
    IPAddressForm, PrefixForm, VLANForm
)
from tenancy.models import Tenant
from django.core.cache import cache
from dcim.api.serializers_.sites import SiteSerializer

__all__ = (
    'SlurpitPlanningViewSet',
    'SlurpitRootView',
    'SlurpitDeviceView'
)

class SlurpitRootView(APIRootView):
    """
    Slurpit API root view
    """
    def get_view_name(self):
        return 'Slurpit'
    

class DeviceViewSet(
        SlurpitViewSet,
        BulkCreateModelMixin,
        BulkDestroyModelMixin,
    ):
    queryset = SlurpitImportedDevice.objects.all()
    serializer_class = SlurpitImportedDeviceSerializer
    filterset_class = SlurpitImportedDeviceFilterSet

    @action(detail=False, methods=['delete'], url_path='delete-all')
    def delete_all(self, request, *args, **kwargs):
        with transaction.atomic():
            Device.objects.select_related('slurpitimporteddevice').update(status=status_decommissioning())
            SlurpitStagedDevice.objects.all().delete()
            SlurpitImportedDevice.objects.filter(mapped_device__isnull=True).delete()

        return Response(status=status.HTTP_204_NO_CONTENT)
    
    @action(detail=False, methods=['delete'], url_path='delete/(?P<hostname>[^/.]+)')
    def delete(self, request, *args, **kwargs):
        hostname_to_delete = kwargs.get('hostname')
        with transaction.atomic():
            to_delete = SlurpitImportedDevice.objects.filter(hostname=hostname_to_delete)
            Device.objects.filter(slurpitimporteddevice__in=to_delete).update(status=status_decommissioning())
            to_delete.filter(mapped_device__isnull=True).delete()
            SlurpitStagedDevice.objects.filter(hostname=hostname_to_delete).delete()

        return Response(status=status.HTTP_204_NO_CONTENT)
    
    def create(self, request):
        sync_sites()
        errors = device_validator(request.data)
        if errors:
            return JsonResponse({'status': 'error', 'errors': errors}, status=400)
        if len(request.data) != 1:
            return JsonResponse({'status': 'error', 'errors': ['List size should be 1']}, status=400)

        start_device_import()
        import_devices(request.data)
        process_import(delete=False)
        
        return JsonResponse({'status': 'success'})
    
    @action(detail=False, methods=['post'],  url_path='sync')
    def sync(self, request):            
        errors = device_validator(request.data)
        if errors:
            return JsonResponse({'status': 'error', 'errors': errors}, status=status.HTTP_400_BAD_REQUEST)
        
        ids = [obj['id'] for obj in request.data]
        hostnames = [obj['hostname'] for obj in request.data]
        SlurpitStagedDevice.objects.filter(Q(hostname__in=hostnames) | Q(slurpit_id__in=ids)).delete()
        import_devices(request.data)        
        return JsonResponse({'status': 'success'})

    @action(detail=False, methods=['post'],  url_path='sync_start')
    def sync_start(self, request):
        sync_sites()
        threshold = timezone.now() - timedelta(days=1)
        SlurpitStagedDevice.objects.filter(createddate__lt=threshold).delete()
        return JsonResponse({'status': 'success'})

    @action(detail=False, methods=['post'],  url_path='sync_end')
    def sync_end(self, request):
        process_import()
        return JsonResponse({'status': 'success'})
    
class SlurpitTestAPIView(SlurpitViewSet):
    queryset = SlurpitImportedDevice.objects.all()
    serializer_class = SlurpitImportedDeviceSerializer
    filterset_class = SlurpitImportedDeviceFilterSet

    permission_classes = [IsAuthenticated]

    @action(detail=False, methods=['get'], url_path='api')
    def api(self, request, *args, **kwargs):    
        return JsonResponse({'status': 'success'})
    
class SlurpitDeviceView(SlurpitViewSet):
    queryset = Device.objects.all()
    serializer_class = DeviceSerializer
    filterset_class = DeviceFilterSet


    @action(detail=False, methods=['get'], url_path='all')
    def all(self, request, *args, **kwargs):
        request_body = []

        devices_array = [get_device_dict(device) for device in Device.objects.all()]

        objs = SlurpitMapping.objects.all()
        
        for device in devices_array:
            row = {}
            for obj in objs:
                target_field = obj.target_field.split('|')[1]
                row[obj.source_field] = str(device[target_field])
            request_body.append(row)


        return JsonResponse({'data': request_body})
    

class SlurpitInterfaceView(SlurpitViewSet):
    queryset = SlurpitInterface.objects.all()

    def get_serializer_class(self):
        return SlurpitInterfaceSerializer
    
    def create(self, request):
        # Validate request Interface data
        errors = interface_validator(request.data)
        if errors:
            return JsonResponse({'status': 'error', 'errors': errors}, status=400)

        try:
            # Get initial values for Interface
            enable_reconcile = True
            initial_obj = SlurpitInterface.objects.filter(name='').values(
                'module', 'type', 'speed', 'label', 'description', 'duplex', 'enable_reconcile', 'ignore_module', 'ignore_type', 'ignore_speed', 'ignore_duplex', 'enabled'
            ).first()
            initial_interface_values = {}
            interface_update_ignore_values = []

            if initial_obj:
                enable_reconcile = initial_obj['enable_reconcile']
                del initial_obj['enable_reconcile']
                initial_interface_values = {**initial_obj}

                for key in initial_interface_values.keys():
                    if key.startswith('ignore_') and initial_interface_values[key]:
                        interface_update_ignore_values.append(key)
            else:
                initial_interface_values = {
                    'type': "other",
                    'label': '',
                    'description': '',
                    'speed': 0,
                    'duplex': None,
                    'module': None,
                    'enabled': True
                }

                # device = None

                # if initial_interface_values['device'] is not None:
                #     device = Device.objects.get(name=initial_interface_values['device'])

                # initial_interface_values['device'] = device

            total_errors = {}
            insert_data = []
            update_data = []
            total_data = []
            duplicates = []
            # Form validation 
            for record in request.data[::-1]:
                unique_interface = f'{record["name"]}/{record["hostname"]}'

                if unique_interface in duplicates:
                    continue
                duplicates.append(unique_interface)
                
                device = None
                try:
                    device = Device.objects.get(name=record['hostname'])
                except: 
                    device = None

                if device is None: 
                    continue
                record['device'] = device
                del record['hostname']

                if 'status' in record:
                    if record['status'] == 'up':
                        record['enabled'] = True
                    else:
                        record['enabled'] = False
                    del record['status']

                if 'description' in record:
                    record['description'] = str(record['description'])
                    
                new_data = {**initial_interface_values, **record}
                total_data.append(new_data)
       
            if enable_reconcile:
                batch_update_qs = []
                batch_insert_qs = []

                for item in total_data:
                    device = None

                    if item['device'] is not None:
                        device = Device.objects.get(name=item['device'])
                        
                    item['device'] = device

                    slurpit_interface_item = SlurpitInterface.objects.filter(name=item['name'], device=item['device'])
                    
                    if slurpit_interface_item:
                        slurpit_interface_item = slurpit_interface_item.first()

                        # Update
                        allowed_fields_with_none = {}
                        allowed_fields = {'duplex', 'label', 'description', 'speed', 'type', 'module', 'enabled'}
                        update = False
                        for field, value in item.items():
                            current = getattr(slurpit_interface_item, field, None)
                            if current != value:
                                if field in allowed_fields and value is not None and value != "":
                                    update = True
                                    setattr(slurpit_interface_item, field, value)
                                if field in allowed_fields_with_none:
                                    update = True
                                    setattr(slurpit_interface_item, field, value)
                        if update:
                            batch_update_qs.append(slurpit_interface_item)
                    else:
                        obj = Interface.objects.filter(name=item['name'], device=item['device'])
                        fields = {'label', 'device', 'module', 'type', 'duplex', 'speed', 'description', 'enabled'}
                        not_null_fields = {'label', 'device', 'module', 'type', 'duplex', 'speed', 'description', 'enabled'}

                        new_interface = {}
                        if obj:
                            obj = obj.first()
                            old_interface = {}

                            for field in fields:
                                field_name = f'ignore_{field}'
                                if field_name in interface_update_ignore_values:
                                    continue
                                old_interface[field] = getattr(obj, field)
                                new_interface[field] = item[field]

                                if field in not_null_fields and (new_interface[field] is None or new_interface[field] == ""):
                                    new_interface[field] = old_interface[field]

                            if new_interface == old_interface:
                                continue
                        else:
                            for field in fields: 
                                new_interface[field] = item[field]

                        batch_insert_qs.append(SlurpitInterface(
                            name = item['name'], 
                            **new_interface
                        ))
                
                count = len(batch_insert_qs)
                offset = 0

                while offset < count:
                    batch_qs = batch_insert_qs[offset:offset + BATCH_SIZE]
                    SlurpitInterface.objects.bulk_create(batch_qs)
                    offset += BATCH_SIZE


                count = len(batch_update_qs)
                offset = 0
                while offset < count:
                    batch_qs = batch_update_qs[offset:offset + BATCH_SIZE]
                    SlurpitInterface.objects.bulk_update(batch_qs, 
                        fields={'label', 'speed', 'type', 'duplex', 'description', 'module'}
                    )
                    offset += BATCH_SIZE

                duplicates = SlurpitInterface.objects.values('name', 'device_id').annotate(count=Count('id')).filter(count__gt=1)
                for duplicate in duplicates:
                    records = SlurpitInterface.objects.filter(name=duplicate['name'], device_id=duplicate['device_id'])                    
                    records.exclude(id=records.first().id).delete()
            else:                
                # Fail case
                for new_data in total_data:
                    obj = Interface()
                    form = InterfaceForm(data=new_data, instance=obj)
                    if form.is_valid() is False:
                        form_errors = form.errors
                        error_list_dict = {}

                        for field, errors in form_errors.items():
                            error_list_dict[field] = list(errors)

                        # Duplicate Interface
                        keys = error_list_dict.keys()
                        
                        if len(keys) ==1 and '__all__' in keys and len(error_list_dict['__all__']) == 1 and error_list_dict['__all__'][0].endswith("already exists."):
                            update_data.append(new_data)
                            continue
                        if '__all__' in keys and len(error_list_dict['__all__']) == 1 and error_list_dict['__all__'][0].endswith("already exists."):
                            del error_list_dict['__all__']
                        
                        error_key = f'{new_data["name"]}({"Global" if new_data["device"] is None else new_data["device"]})'
                        total_errors[error_key] = error_list_dict

                        return JsonResponse({'status': 'error', 'errors': total_errors}, status=400)
                    else:
                        insert_data.append(new_data)

                # Batch Insert
                count = len(insert_data)
                offset = 0
                while offset < count:
                    batch_qs = insert_data[offset:offset + BATCH_SIZE]
                    to_import = []        
                    for interface_item in batch_qs:
                        filtered_interface_item = {k: v for k, v in interface_item.items() if not k.startswith('ignore_')}
                        to_import.append(Interface(**filtered_interface_item))
                    Interface.objects.bulk_create(to_import)
                    offset += BATCH_SIZE
                
                
                # Batch Update
                batch_update_qs = []
                for update_item in update_data:
                    item = Interface.objects.get(name=update_item['name'], device=update_item['device'])
                    
                    # Update
                    allowed_fields_with_none = {}
                    allowed_fields = {'duplex', 'label', 'speed', 'type', 'description', 'module'}

                    for field, value in update_item.items():
                        ignore_field = f'ignore_{field}'
                        if ignore_field in interface_update_ignore_values:
                            continue 

                        if field in allowed_fields and value is not None and value != "":
                            setattr(item, field, value)
                        if field in allowed_fields_with_none:
                            setattr(item, field, value)

                    batch_update_qs.append(item)

                
                count = len(batch_update_qs)
                offset = 0
                while offset < count:
                    batch_qs = batch_update_qs[offset:offset + BATCH_SIZE]
                    to_import = []        
                    for interface_item in batch_qs:
                        to_import.append(interface_item)

                    Interface.objects.bulk_update(to_import, 
                        fields={'label', 'speed', 'type', 'duplex', 'description', 'module'}
                    )
                    offset += BATCH_SIZE


            return JsonResponse({'status': 'success'})
        except Exception as e:
            return JsonResponse({'status': 'errors', 'errors': f"Process threw error: {e} - {traceback.format_exc()}"}, status=400)
        
class SlurpitIPAMView(SlurpitViewSet):
    queryset = IPAddress.objects.all()
    
    def get_serializer_class(self):
        return SlurpitInitIPAddressSerializer
    
    def create(self, request):
        # Validate request IPAM data
        errors = ipam_validator(request.data)
        if errors:
            return JsonResponse({'status': 'error', 'errors': errors}, status=400)

        vrf = None
        tenant = None

        try:
            # Get initial values for IPAM
            enable_reconcile = True
            initial_obj = SlurpitInitIPAddress.objects.filter(address=None).values(
                'status', 'vrf', 'tenant', 'role', 'enable_reconcile', 'description', 'ignore_status', 'ignore_vrf', 'ignore_tenant', 'ignore_role', 'ignore_description'
            ).first()

            initial_ipaddress_values = {}
            ipaddress_update_ignore_values = []
            vrf = None
            tenant = None
            if initial_obj:
                enable_reconcile = initial_obj['enable_reconcile']
                del initial_obj['enable_reconcile']
                initial_ipaddress_values = {**initial_obj}

                obj = SlurpitInitIPAddress.objects.filter(address=None).get()

                if initial_ipaddress_values['vrf'] is not None:
                    vrf = VRF.objects.get(pk=initial_ipaddress_values['vrf'])
                if initial_ipaddress_values['tenant'] is not None:
                    tenant = Tenant.objects.get(pk=initial_ipaddress_values['tenant'])

                initial_ipaddress_values['vrf'] = vrf
                initial_ipaddress_values['tenant'] = tenant

                for key in initial_ipaddress_values.keys():
                    if key.startswith('ignore_') and initial_ipaddress_values[key]:
                        ipaddress_update_ignore_values.append(key)
                
            else:
                initial_ipaddress_values['vrf'] = None
                initial_ipaddress_values['tenant'] = None
                initial_ipaddress_values['role'] = ''
                initial_ipaddress_values['description'] = ''
                initial_ipaddress_values['status'] = 'active'

            total_errors = {}
            insert_ips = []
            update_ips = []
            total_ips = []

            duplicates = []
            # Form validation 
            for record in request.data[::-1]:
                unique_ipaddress = f'{record["address"]}'

                if unique_ipaddress in duplicates:
                    continue

                if unique_ipaddress.endswith("/32"):
                    unique_ipaddress = unique_ipaddress[:-3]

                if '/' not in unique_ipaddress:
                    prefix = Prefix.objects.filter(prefix__net_contains=unique_ipaddress).order_by('-prefix').first()

                    if prefix:
                        unique_ipaddress = f'{unique_ipaddress}/{prefix.prefix.prefixlen}'
                    else:
                        unique_ipaddress = f'{unique_ipaddress}/32'

                record['address'] = unique_ipaddress
                
                duplicates.append(unique_ipaddress)

                new_data = {**initial_ipaddress_values, **record}
                total_ips.append(new_data)



            if enable_reconcile:
                batch_update_qs = []
                batch_insert_qs = []

                for item in total_ips:

                    slurpit_ipaddress_item = SlurpitInitIPAddress.objects.filter(address=item['address'], vrf=item['vrf'])
                    
                    if slurpit_ipaddress_item:
                        slurpit_ipaddress_item = slurpit_ipaddress_item.first()

                        allowed_fields_with_none = {'status'}
                        allowed_fields = {'role', 'tenant', 'dns_name', 'description'}
                        update = False
                        for field, value in item.items():
                            current = getattr(slurpit_ipaddress_item, field, None)
                            if current != value:
                                if field in allowed_fields and value is not None and value != "":
                                    update = True
                                    setattr(slurpit_ipaddress_item, field, value)
                                if field in allowed_fields_with_none:
                                    update = True
                                    setattr(slurpit_ipaddress_item, field, value)
                        if update:
                            batch_update_qs.append(slurpit_ipaddress_item)
                    else:
                        obj = IPAddress.objects.filter(address=item['address'], vrf=vrf)
                        fields = ['status', 'role', 'description', 'tenant', 'dns_name']
                        not_null_fields = {'role', 'description', 'tenant', 'dns_name'}
                        new_ipaddress = {}

                        if obj:
                            obj = obj.first()
                            old_ipaddress = {}
                            
                            for field in fields:
                                field_name = f'ignore_{field}'
                                if field_name in ipaddress_update_ignore_values:
                                    continue
                                old_ipaddress[field] =   getattr(obj, field)
                                new_ipaddress[field] = item[field]

                                if field in not_null_fields and (new_ipaddress[field] is None or new_ipaddress[field] == ""):
                                    new_ipaddress[field] = old_ipaddress[field]

                            if new_ipaddress == old_ipaddress:
                                continue
                        else:
                            for field in fields:
                                new_ipaddress[field] = item[field]
                        
                        obj = SlurpitInitIPAddress(
                            address = item['address'], 
                            vrf = vrf,
                            **new_ipaddress
                        )

                        batch_insert_qs.append(obj)
                
                count = len(batch_insert_qs)
                offset = 0

                while offset < count:
                    batch_qs = batch_insert_qs[offset:offset + BATCH_SIZE]
                    created_items = SlurpitInitIPAddress.objects.bulk_create(batch_qs)
                    offset += BATCH_SIZE

                count = len(batch_update_qs)
                offset = 0
                while offset < count:
                    batch_qs = batch_update_qs[offset:offset + BATCH_SIZE]
                    SlurpitInitIPAddress.objects.bulk_update(batch_qs, fields={'status', 'role', 'tenant', 'dns_name', 'description'})
                    offset += BATCH_SIZE

                duplicates = SlurpitInitIPAddress.objects.values('address', 'vrf').annotate(count=Count('id')).filter(count__gt=1)
                for duplicate in duplicates:
                    records = SlurpitInitIPAddress.objects.filter(address=duplicate['address'], vrf=duplicate['vrf'])                    
                    records.exclude(id=records.first().id).delete()
                
            else:
                # Fail case
                for new_data in total_ips:
                    obj = IPAddress()
                    form = IPAddressForm(data=new_data, instance=obj)
                    if form.is_valid() is False:
                        form_errors = form.errors
                        error_list_dict = {}

                        for field, errors in form_errors.items():
                            error_list_dict[field] = list(errors)

                        # Duplicate IP Address
                        keys = error_list_dict.keys()
                        
                        if len(keys) ==1 and 'address' in keys and len(error_list_dict['address']) == 1 and error_list_dict['address'][0].startswith("Duplicate"):
                            update_ips.append(new_data)
                            continue
                        if 'address' in keys and len(error_list_dict['address']) == 1 and error_list_dict['address'][0].startswith("Duplicate"):
                            del error_list_dict['address']
                        
                        error_key = f'{new_data["address"]}({"Global" if new_data["vrf"] is None else new_data["vrf"]})'
                        total_errors[error_key] = error_list_dict

                        return JsonResponse({'status': 'error', 'errors': total_errors}, status=400)
                    else:
                        insert_ips.append(new_data)

                # Batch Insert
                count = len(insert_ips)
                offset = 0
                while offset < count:
                    batch_qs = insert_ips[offset:offset + BATCH_SIZE]
                    to_import = []        
                    for ipaddress_item in batch_qs:
                        filtered_ipaddress_item = {k: v for k, v in ipaddress_item.items() if not k.startswith('ignore_')}
                        to_import.append(IPAddress(**filtered_ipaddress_item))

                    IPAddress.objects.bulk_create(to_import)
                    offset += BATCH_SIZE
                
                # Batch Update
                batch_update_qs = []
                for update_item in update_ips:
                    item = IPAddress.objects.get(address=update_item['address'], vrf=update_item['vrf'])

                    # Update
                    allowed_fields_with_none = {'status'}
                    allowed_fields = {'role', 'tenant', 'dns_name', 'description'}

                    for field, value in update_item.items():
                        ignore_field = f'ignore_{field}'
                        if ignore_field in ipaddress_update_ignore_values:
                            continue 

                        if field in allowed_fields and value is not None and value != "":
                            setattr(item, field, value)
                        if field in allowed_fields_with_none:
                            setattr(item, field, value)

                    batch_update_qs.append(item)

                count = len(batch_update_qs)
                offset = 0
                while offset < count:
                    batch_qs = batch_update_qs[offset:offset + BATCH_SIZE]

                    IPAddress.objects.bulk_update(batch_qs, fields={'status', 'role', 'tenant', 'dns_name', 'description'})
                    offset += BATCH_SIZE


            return JsonResponse({'status': 'success'})
        except Exception as e:
            return JsonResponse({'status': 'error', 'errors': str(e)}, status=400)
        


class SlurpitPrefixView(SlurpitViewSet):
    queryset = SlurpitPrefix.objects.all()

    def get_serializer_class(self):
        return SlurpitPrefixSerializer
    
    def create(self, request):
        # Validate request prefix data
        errors = prefix_validator(request.data)
        if errors:
            return JsonResponse({'status': 'error', 'errors': errors}, status=400)

        vrf = None
        tenant = None
        vlan = None
        role = None
        site = None
        scope = None
        scope_id = None
        scope_type = None
        site_group = None
        location = None
        region = None
            
        try:
            # Get initial values for prefix
            enable_reconcile = True
            initial_obj = SlurpitPrefix.objects.filter(prefix=None).values(
                'status', 'vrf', 'role', 'vlan', 'tenant', 'enable_reconcile', 'description', 'ignore_status', 'ignore_vrf', 'ignore_role', 'ignore_site', 'ignore_vlan', 'ignore_tenant', 'ignore_description', 'scope_id', 'scope_type', '_site', '_site_group', '_region', '_location'
            ).first()
            initial_prefix_values = {}
            prefix_update_ignore_values = []

            if initial_obj:
                enable_reconcile = initial_obj['enable_reconcile']
                del initial_obj['enable_reconcile']
                initial_prefix_values = {**initial_obj}

                if initial_prefix_values['vrf'] is not None:
                    vrf = VRF.objects.get(pk=initial_prefix_values['vrf'])
                if initial_prefix_values['tenant'] is not None:
                    tenant = Tenant.objects.get(pk=initial_prefix_values['tenant'])
                if initial_prefix_values['vlan'] is not None:
                    vlan = VLAN.objects.get(pk=initial_prefix_values['vlan'])
                if initial_prefix_values['role'] is not None:
                    role = Role.objects.get(pk=initial_prefix_values['role'])
                # if initial_prefix_values['scope'] is not None:
                #     scope = Scope.objects.get(pk=initial_prefix_values['scope'])
                # if initial_prefix_values['scope_type'] is not None:
                #     scope_type = ScopeType.objects.get(pk=initial_prefix_values['scope_type'])
                if initial_prefix_values['_site'] is not None:
                    site = Site.objects.get(pk=initial_prefix_values['_site'])
                if initial_prefix_values['_site_group'] is not None:
                    site_group = SiteGroup.objects.get(pk=initial_prefix_values['_site_group'])
                if initial_prefix_values['_location'] is not None:
                    location = Location.objects.get(pk=initial_prefix_values['_location'])
                if initial_prefix_values['_region'] is not None:
                    region = Region.objects.get(pk=initial_prefix_values['_region'])

                initial_prefix_values['vrf'] = vrf
                initial_prefix_values['tenant'] = tenant
                initial_prefix_values['vlan'] = vlan
                initial_prefix_values['role'] = role
                initial_prefix_values['scope_id'] = scope_id
                initial_prefix_values['scope_type'] = scope_type
                initial_prefix_values['_site'] = site
                initial_prefix_values['_site_group'] = site_group
                initial_prefix_values['_location'] = location
                initial_prefix_values['_region'] = region

                for key in initial_prefix_values.keys():
                    if key.startswith('ignore_') and initial_prefix_values[key]:
                        prefix_update_ignore_values.append(key)

            else:
                initial_prefix_values = {
                    'status': 'active',
                    'vrf': None,
                    'scope_id': None,
                    'scope_type': None,
                    '_site': None,
                    '_site_group': None,
                    '_location': None,
                    '_region': None,
                    'tenant': None,
                    'vlan': None,
                    'role': None,
                    'description': ''
                }

            total_errors = {}
            insert_data = []
            update_data = []
            total_data = []

            duplicates = []
            # Form validation 
            for record in request.data[::-1]:
                unique_prefix = f'{record["prefix"]}'

                if unique_prefix in duplicates:
                    continue
                duplicates.append(unique_prefix)

                if 'vrf' in record and len(record['vrf']) > 0:
                    vrf = VRF.objects.filter(name=record['vrf'])
                    if vrf:
                        record['vrf'] = vrf.first()
                    else:
                        vrf = VRF.objects.create(name=record['vrf'])
                        record['vrf'] = vrf
                else:
                    if 'vrf' in record:
                        del record['vrf']

                new_data = {**initial_prefix_values, **record}
                total_data.append(new_data)
        
            if enable_reconcile:
                batch_update_qs = []
                batch_insert_qs = []

                for item in total_data:

                    slurpit_prefix_item = SlurpitPrefix.objects.filter(prefix=item['prefix'], vrf=item['vrf'])
                    
                    if slurpit_prefix_item:
                        slurpit_prefix_item = slurpit_prefix_item.first()

                        allowed_fields_with_none = {'status'}
                        allowed_fields = {'role', 'tenant', 'vlan', 'vrf', 'description', 'scope_id', 'scope_type', '_site', '_site_group', '_location', '_region'}
                        update = False
                        for field, value in item.items():
                            current = getattr(slurpit_prefix_item, field, None)
                            if current != value:
                                if field in allowed_fields and value is not None and value != "":
                                    update = True
                                    setattr(slurpit_prefix_item, field, value)
                                if field in allowed_fields_with_none:
                                    update = True
                                    setattr(slurpit_prefix_item, field, value)
                        if update:
                            batch_update_qs.append(slurpit_prefix_item)
                    else:
                        obj = Prefix.objects.filter(prefix=item['prefix'], vrf=item['vrf'])
                        
                        fields = {'status', 'vrf', 'vlan', 'tenant', 'role', 'description', 'scope_id', 'scope_type', '_site', '_site_group', '_location', '_region'}
                        not_null_fields = {'vlan', 'tenant', 'role', 'description'}
                        
                        new_prefix = {}

                        if obj:
                            obj = obj.first()
                            old_prefix = {}
                            
                            for field in fields:
                                field_name = f'ignore_{field}'
                                if field_name in prefix_update_ignore_values:
                                    continue
                                old_prefix[field] = getattr(obj, field)
                                new_prefix[field] = item[field]

                                if field in not_null_fields and (new_prefix[field] is None or new_prefix[field] == ""):
                                    new_prefix[field] = old_prefix[field]

                            if new_prefix == old_prefix:
                                continue
                        else:
                            for field in fields:
                                new_prefix[field] = item.get(field, None)

                        batch_insert_qs.append(SlurpitPrefix(
                            prefix = item['prefix'],
                            **new_prefix
                        ))
                
                count = len(batch_insert_qs)
                offset = 0

                while offset < count:
                    batch_qs = batch_insert_qs[offset:offset + BATCH_SIZE]
                    SlurpitPrefix.objects.bulk_create(batch_qs)
                    offset += BATCH_SIZE


                count = len(batch_update_qs)
                offset = 0
                while offset < count:
                    batch_qs = batch_update_qs[offset:offset + BATCH_SIZE]
                    SlurpitPrefix.objects.bulk_update(batch_qs, fields={'description', 'vrf', 'tenant', 'status', 'vlan', 'role', 'scope_id', 'scope_type', '_site', '_site_group', '_location', '_region'})
                    offset += BATCH_SIZE

                duplicates = SlurpitPrefix.objects.values('prefix', 'vrf').annotate(count=Count('id')).filter(count__gt=1)
                for duplicate in duplicates:
                    records = SlurpitPrefix.objects.filter(prefix=duplicate['prefix'], vrf=duplicate['vrf'])                    
                    records.exclude(id=records.first().id).delete()

            else:                
                # Fail case
                for new_data in total_data:
                    obj = Prefix()
                    form = PrefixForm(data=new_data, instance=obj)
                    if form.is_valid() is False:
                        form_errors = form.errors
                        error_list_dict = {}

                        for field, errors in form_errors.items():
                            error_list_dict[field] = list(errors)

                        # Duplicate Prefix
                        keys = error_list_dict.keys()
                        
                        if len(keys) ==1 and 'prefix' in keys and len(error_list_dict['prefix']) == 1 and error_list_dict['prefix'][0].startswith("Duplicate"):
                            update_data.append(new_data)
                            continue
                        if 'prefix' in keys and len(error_list_dict['prefix']) == 1 and error_list_dict['prefix'][0].startswith("Duplicate"):
                            del error_list_dict['prefix']
                        
                        error_key = f'{new_data["prefix"]}({"Global" if new_data["vrf"] is None else new_data["vrf"]})'
                        total_errors[error_key] = error_list_dict

                        return JsonResponse({'status': 'error', 'errors': total_errors}, status=400)
                    else:
                        insert_data.append(new_data)

                # Batch Insert
                count = len(insert_data)
                offset = 0
                while offset < count:
                    batch_qs = insert_data[offset:offset + BATCH_SIZE]
                    to_import = []        
                    for prefix_item in batch_qs:
                        filtered_prefix_item = {k: v for k, v in prefix_item.items() if not k.startswith('ignore_')}
                        to_import.append(Prefix(**filtered_prefix_item))
                    Prefix.objects.bulk_create(to_import)
                    offset += BATCH_SIZE
                
                
                # Batch Update
                batch_update_qs = []
                for update_item in update_data:
                    item = Prefix.objects.get(prefix=update_item['prefix'], vrf=update_item['vrf'])
                    
                    # Update
                    allowed_fields_with_none = {'status'}
                    allowed_fields = {'role', 'tenant', 'vlan', 'description', 'vrf', 'scope_id', 'scope_type', '_site', '_site_group', '_location', '_region'}

                    for field, value in update_item.items():
                        ignore_field = f'ignore_{field}'
                        if ignore_field in prefix_update_ignore_values:
                            continue 
                        
                        if field in allowed_fields and value is not None and value != "":
                            setattr(item, field, value)
                        if field in allowed_fields_with_none:
                            setattr(item, field, value)
                    
                    batch_update_qs.append(item)

                
                count = len(batch_update_qs)
                offset = 0
                while offset < count:
                    batch_qs = batch_update_qs[offset:offset + BATCH_SIZE]
                    to_import = []        
                    for prefix_item in batch_qs:
                        to_import.append(prefix_item)

                    Prefix.objects.bulk_update(to_import, 
                        fields={
                            'description', 'vrf', 'tenant', 'status', 'vlan', 'role', 'scope_id', 'scope_type', '_site', '_site_group', '_location', '_region'
                        }
                    )
                    offset += BATCH_SIZE


            return JsonResponse({'status': 'success'})
        except Exception as e:
            return JsonResponse({'status': 'errors', 'errors': f"Process threw error: {e} - {traceback.format_exc()}"}, status=400)

    @action(detail=False, methods=['get'], url_path='all')
    def all(self, request, *args, **kwargs):
        prefixes = Prefix.objects.filter(tags__name="slurpit")
        serializer = PrefixSerializer(prefixes, many=True, context={'request': request})
        
        prefixes = []
        for prefix in serializer.data:
            prefixes.append(f'{prefix["prefix"]}')
        return JsonResponse(prefixes, safe=False)

class SlurpitVLANView(SlurpitViewSet):
    queryset = SlurpitVLAN.objects.all()

    def get_serializer_class(self):
        return SlurpitVLANSerializer
    
    def create(self, request):
        # Validate request vlan data
        errors = vlan_validator(request.data)
        if errors:
            return JsonResponse({'status': 'error', 'errors': errors}, status=400)

        tenant = None
        role = None
            
        try:
            # Get initial values for vlan
            enable_reconcile = True
            initial_obj = SlurpitVLAN.objects.filter(name='').values(
                'status', 'role', 'tenant', 'enable_reconcile', 'description', 
                'ignore_status', 'ignore_role', 'ignore_tenant', 'ignore_description'
            ).first()
            initial_vlan_values = {}
            vlan_update_ignore_values = []

            if initial_obj:
                enable_reconcile = initial_obj['enable_reconcile']
                del initial_obj['enable_reconcile']
                initial_vlan_values = {**initial_obj}

                if initial_vlan_values['tenant'] is not None:
                    tenant = Tenant.objects.get(pk=initial_vlan_values['tenant'])
                if initial_vlan_values['role'] is not None:
                    role = Role.objects.get(pk=initial_vlan_values['role'])

                initial_vlan_values['tenant'] = tenant
                initial_vlan_values['role'] = role

                for key in initial_vlan_values.keys():
                    if key.startswith('ignore_') and initial_vlan_values[key]:
                        vlan_update_ignore_values.append(key)

            else:
                initial_vlan_values = {
                    'status': 'active',
                    'tenant': None,
                    'role': None,
                    'description': ''
                }

            total_errors = {}
            insert_data = []
            update_data = []
            total_data = []

            duplicates = []
            # Form validation 
            for record in request.data[::-1]:
                unique_group_name = f'{record["hostname"]}_{record["vlan_name"]}'
                unique_group_id = f'{record["hostname"]}_{record["vlan_id"]}'
                if unique_group_name in duplicates or unique_group_id in duplicates:
                    continue
                duplicates.append(unique_group_name)
                duplicates.append(unique_group_id)


                new_data = {
                    **initial_vlan_values, 
                    "vid": record['vlan_id'],
                    "name": record['vlan_name'],
                    "hostname": record['hostname']
                }
                # Get VLANGroup ID 
                group = VLANGroup.objects.filter(name=record['hostname'])
                if group:
                    group = group.first()
                    new_data['group'] = group
                total_data.append(new_data)
        
            if enable_reconcile:
                batch_update_qs = []
                batch_insert_qs = []

                for item in total_data:
                    slurpit_vlan_item = SlurpitVLAN.objects.filter(name=item['name'], group=item['hostname'])
                    if not slurpit_vlan_item:
                        slurpit_vlan_item = SlurpitVLAN.objects.filter(vid=item['vid'], group=item['hostname'])

                    if slurpit_vlan_item:
                        slurpit_vlan_item = slurpit_vlan_item.first()

                        allowed_fields_with_none = {'status'}
                        allowed_fields = {'role', 'tenant', 'description'}
                        update = False

                        for field, value in item.items():
                            current = getattr(slurpit_vlan_item, field, None)
                            if current != value:
                                if field in allowed_fields and value is not None and value != "":
                                    update = True
                                    setattr(slurpit_vlan_item, field, value)
                                if field in allowed_fields_with_none:
                                    update = True
                                    setattr(slurpit_vlan_item, field, value)
                        if update:
                            batch_update_qs.append(slurpit_vlan_item)
                    else:
                        obj = VLAN.objects.filter(name=item['name'], group__name=item['hostname'])
                        if not obj:
                            obj = VLAN.objects.filter(vid=item['vid'], group__name=item['hostname'])

                        fields = {'status', 'tenant', 'role', 'description', 'name'}
                        not_null_fields = {'tenant', 'role', 'description'}
                        
                        new_vlan = {}

                        if obj:
                            obj = obj.first()
                            old_vlan = {}
                            
                            for field in fields:
                                field_name = f'ignore_{field}'
                                if field_name in vlan_update_ignore_values:
                                    continue
                                old_vlan[field] = getattr(obj, field)
                                new_vlan[field] = item[field]

                                if field in not_null_fields and (new_vlan[field] is None or new_vlan[field] == ""):
                                    new_vlan[field] = old_vlan[field]

                            if new_vlan == old_vlan:
                                continue
                        else:
                            for field in fields:
                                new_vlan[field] = item[field]

                        batch_insert_qs.append(SlurpitVLAN(
                            # name = item['name'],
                            vid = item['vid'],
                            group = item['hostname'],
                            **new_vlan
                        ))
                
                count = len(batch_insert_qs)
                offset = 0

                while offset < count:
                    batch_qs = batch_insert_qs[offset:offset + BATCH_SIZE]
                    SlurpitVLAN.objects.bulk_create(batch_qs)
                    offset += BATCH_SIZE


                count = len(batch_update_qs)
                offset = 0
                while offset < count:
                    batch_qs = batch_update_qs[offset:offset + BATCH_SIZE]
                    SlurpitVLAN.objects.bulk_update(batch_qs, fields={'description', 'tenant', 'status', 'role'})
                    offset += BATCH_SIZE

                duplicates = SlurpitVLAN.objects.values('name', 'group').annotate(count=Count('id')).filter(count__gt=1)
                for duplicate in duplicates:
                    records = SlurpitVLAN.objects.filter(name=duplicate['name'], group=duplicate['hostname'])                    
                    records.exclude(id=records.first().id).delete()

                duplicates = SlurpitVLAN.objects.values('vid', 'group').annotate(count=Count('id')).filter(count__gt=1)
                for duplicate in duplicates:
                    records = SlurpitVLAN.objects.filter(vid=duplicate['vid'], group=duplicate['hostname'])                    
                    records.exclude(id=records.first().id).delete()

            else:                
                # Fail case
                for new_data in total_data:
                    obj = VLAN()
                    form = VLANForm(data=new_data, instance=obj)
                    if form.is_valid() is False:
                        form_errors = form.errors
                        error_list_dict = {}

                        for field, errors in form_errors.items():
                            error_list_dict[field] = list(errors)

                        # Duplicate VLAN
                        keys = error_list_dict.keys()

                        if len(keys) ==1 and '__all__' in keys:
                            flag = True
                            for errorItem in error_list_dict['__all__']:
                                if not errorItem.endswith("already exists."):
                                    flag = False
                                    break
                            if flag:
                                update_data.append(new_data)
                                continue
                        if '__all__' in keys:
                            del error_list_dict['__all__']
                        
                        error_key = f'{new_data["name"]}({new_data["vid"]})'
                        total_errors[error_key] = error_list_dict

                        return JsonResponse({'status': 'error', 'errors': total_errors}, status=400)
                    else:
                        insert_data.append(new_data)

                # Batch Insert
                count = len(insert_data)
                offset = 0
                while offset < count:
                    batch_qs = insert_data[offset:offset + BATCH_SIZE]
                    to_import = []        
                    for vlan_item in batch_qs:
                        filtered_vlan_item = {k: v for k, v in vlan_item.items() if not k.startswith('ignore_')}
                        if not 'group' in filtered_vlan_item:
                            vlan_group = VLANGroup.objects.create(name=filtered_vlan_item['hostname'],slug=filtered_vlan_item['hostname'])
                            filtered_vlan_item['group'] = vlan_group
                        del filtered_vlan_item['hostname']
                        to_import.append(VLAN(**filtered_vlan_item))
                    VLAN.objects.bulk_create(to_import)
                    offset += BATCH_SIZE
                
                # Batch Update
                batch_update_qs = []
                for update_item in update_data:
                    if not 'group' in update_item:
                        vlan_group = VLANGroup.objects.create(name=update_item['hostname'],slug=update_item['hostname'])
                        update_item['group'] = vlan_group

                    item = VLAN.objects.filter(name=update_item['name'], group=update_item['group'])
                    if not item:
                        item = VLAN.objects.filter(vid=update_item['vid'], group=update_item['group'])

                    if item:
                        item = item.first()
                    # Update
                    allowed_fields_with_none = {'status'}
                    allowed_fields = {'role', 'tenant', 'description', 'name'}

                    for field, value in update_item.items():
                        ignore_field = f'ignore_{field}'
                        if ignore_field in vlan_update_ignore_values:
                            continue 
                        
                        if field in allowed_fields and value is not None and value != "":
                            setattr(item, field, value)
                        if field in allowed_fields_with_none:
                            setattr(item, field, value)
                    
                    batch_update_qs.append(item)

                
                count = len(batch_update_qs)
                offset = 0
                while offset < count:
                    batch_qs = batch_update_qs[offset:offset + BATCH_SIZE]
                    to_import = []        
                    for vlan_item in batch_qs:
                        to_import.append(vlan_item)

                    VLAN.objects.bulk_update(to_import, 
                        fields={
                            'description', 'tenant', 'status', 'role', 'name'
                        }
                    )
                    offset += BATCH_SIZE


            return JsonResponse({'status': 'success'})
        except Exception as e:
            return JsonResponse({'status': 'errors', 'errors': str(e)}, status=400)

class SlurpitPlanningViewSet(
        SlurpitViewSet
    ):
    queryset = SlurpitPlanning.objects.all()
    serializer_class = SlurpitPlanningSerializer
    filterset_class = SlurpitPlanningFilterSet
    
    def get_queryset(self):
        if self.request.method == 'GET':
            # Customize this queryset to suit your requirements for GET requests
            return SlurpitPlanning.objects.filter(selected=True)
        # For other methods, use the default queryset
        return self.queryset
    
class SlurpitSiteView(SlurpitViewSet):
    queryset = Site.objects.all()
    
    def get_serializer_class(self):
        return SiteSerializer
    
    def create(self, request):
        try:
            create_sites(request.data[::-1])
        except Exception as e:
            return JsonResponse({'status': 'errors', 'errors': str(e)}, status=400)

        return JsonResponse({'status': 'success'})
        



# Generated by Django 4.2.9 on 2024-04-19 08:05

import django.core.validators
from django.db import migrations, models
import ipam.fields


class Migration(migrations.Migration):

    dependencies = [
        ('slurpit_netbox', '0007_slurpitinitipaddress'),
    ]

    operations = [
        migrations.AddField(
            model_name='slurpitinitipaddress',
            name='address',
            field=ipam.fields.IPAddressField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='slurpitinitipaddress',
            name='dns_name',
            field=models.CharField(blank=True, max_length=255, null=True, validators=[django.core.validators.RegexValidator(code='invalid', message='Only alphanumeric characters, asterisks, hyphens, periods, and underscores are allowed in DNS names', regex='^([0-9A-Za-z_-]+|\\*)(\\.[0-9A-Za-z_-]+)*\\.?$')]),
        ),
    ]



import requests

from datetime import datetime

from django.core.exceptions import ObjectDoesNotExist
from django.db import connection, transaction
from django.db.models import QuerySet, F, OuterRef, Subquery
from django.utils import timezone
from django.db.models.expressions import RawSQL
from django.utils.text import slugify

from . import get_config
from .models import SlurpitImportedDevice, SlurpitStagedDevice, ensure_slurpit_tags, SlurpitSetting, SlurpitPlanning, SlurpitSnapshot
from .management.choices import *
from .references import base_name, plugin_type, custom_field_data_name
from .references.generic import get_default_objects, status_inventory, status_offline, get_create_dcim_objects, set_device_custom_fields, status_active
from .references.imports import *
from dcim.models import Interface, Site
from ipam.models import IPAddress, Prefix

BATCH_SIZE = 512
columns = ('slurpit_id', 'disabled', 'hostname', 'fqdn', 'ipv4', 'device_os', 'device_type', 'brand', "serial", "os_version", "snmp_uptime", 'createddate', 'changeddate', 'site')

import re

ti = 50
def slug_string(input_string, ti):
    # Remove all characters except hyphens, dots, word characters, and whitespace
    cleaned_string = re.sub(r'[^\-.\w\s]', '', input_string)
    
    # Remove leading and trailing spaces or dots
    cleaned_string = re.sub(r'^[\s.]+|[\s.]+$', '', cleaned_string)
    
    # Replace sequences of hyphens, dots, or whitespace with a single hyphen
    cleaned_string = re.sub(r'[-.\s]+', '-', cleaned_string)
    
    # Convert to lowercase and truncate to length ti
    result_string = cleaned_string.lower()[:ti]
    
    return result_string

def get_devices(offset):
    try:
        setting = SlurpitSetting.objects.get()
        uri_base = setting.server_url
        headers = {
                        'authorization': setting.api_key,
                        'useragent': f"{plugin_type}/requests",
                        'accept': 'application/json'
                    }
        uri_devices = f"{uri_base}/api/devices?offset={offset}&limit={BATCH_SIZE}"
        r = requests.get(uri_devices, headers=headers, timeout=15, verify=False)
        # r.raise_for_status()
        data = r.json()
        return data, ""
    except ObjectDoesNotExist as e:
        log_message = "Api key not configured in settings"
        return None, log_message
    except Exception as e:
        log_message = "Please confirm the Slurp'it server is running and reachable."
        return None, log_message

def start_device_import():
    with connection.cursor() as cursor:
        cursor.execute(f"truncate {SlurpitStagedDevice._meta.db_table} cascade")

def format_address(street, number, zipcode, country):
    return f"{street} {number} {zipcode} {country}"

def create_sites(data):
    for item in data:
        if 'sitename' not in item or item['sitename'] == "":
            continue

        # First, format the address
        address = format_address(item['street'], item['number'], item['zipcode'], item['country'])
        
        if item['latitude'] == '':
            item['latitude'] = None
        
        if item['longitude'] == '':
            item['longitude'] = None

        if item['status'] == '1':
            status = 'active'
        else:
            status = 'retired'

        # Prepare data for the Site instance
        site_data = {
            'description': item['description'],
            'longitude': item['longitude'],
            'latitude': item['latitude'],
            'slug': slug_string(item['sitename'], ti),
            'status': status,
            'physical_address': address,
            'shipping_address': address,
        }

        # Update if exists, create if not
        site, created = Site.objects.update_or_create(
            name=item['sitename'],  # Field to match for finding the record
            defaults=site_data       # Fields to update or set if the object is created
        )

        if created:
            print(f"Created new site with name {item['sitename']}")
        else:
            print(f"Updated existing site with name {item['sitename']}")

def sync_sites():
    try:
        setting = SlurpitSetting.objects.get()
        uri_base = setting.server_url
        headers = {
                        'authorization': setting.api_key,
                        'useragent': f"{plugin_type}/requests",
                        'accept': 'application/json'
                    }
        uri_sites = f"{uri_base}/api/sites"
        r = requests.get(uri_sites, headers=headers, timeout=15, verify=False)
        data = r.json()

        # Import Slurpit Sites to NetBox
        create_sites(data)

        return data, ""
    except ObjectDoesNotExist as e:
        log_message = "Api key not configured in settings"
        return None, log_message
    except Exception as e:
        print(e)
        log_message = "Please confirm the Slurp'it server is running and reachable."
        return None, log_message

def import_devices(devices):
    to_insert = []
    for device in devices:
        # if device.get('disabled') == '1':
        #     continue
        if device.get('device_type') is None:
            continue
        device['slurpit_id'] = device.pop('id')
        
        try:
            device['createddate'] = timezone.make_aware(datetime.strptime(device['createddate'], '%Y-%m-%d %H:%M:%S'), timezone.get_current_timezone())
            device['changeddate'] = timezone.make_aware(datetime.strptime(device['changeddate'], '%Y-%m-%d %H:%M:%S'), timezone.get_current_timezone())          
        except ValueError:
            continue
        to_insert.append(SlurpitStagedDevice(**{key: value for key, value in device.items() if key in columns}))
    SlurpitStagedDevice.objects.bulk_create(to_insert)
    

def process_import(delete=True):
    if delete:
        handle_parted()
    handle_changed()
    handle_new_comers()
    
def run_import():
    devices = get_devices()
    if devices is not None:
        start_device_import()
        import_devices(devices)
        process_import()
        return 'done'
    else:
        return 'none'


def handle_parted():
    parted_qs = SlurpitImportedDevice.objects.exclude(
        slurpit_id__in=SlurpitStagedDevice.objects.values('slurpit_id')
    )
    
    count = 0
    for device in parted_qs:
        if device.mapped_device is None:
            device.delete()
        # elif device.mapped_device.status == status_offline():
        #     continue
        # else:
        #     device.mapped_device.status=status_offline()
        #     device.mapped_device.save()
        count += 1
    

def handle_new_comers():
    unattended = get_config('unattended_import')
    
    qs = SlurpitStagedDevice.objects.exclude(
        slurpit_id__in=SlurpitImportedDevice.objects.values('slurpit_id')
    )

    offset = 0
    count = len(qs)

    while offset < count:
        batch_qs = qs[offset:offset + BATCH_SIZE]
        to_import = []        
        for device in batch_qs:
            to_import.append(get_from_staged(device, unattended))
        SlurpitImportedDevice.objects.bulk_create(to_import, ignore_conflicts=True)
        offset += BATCH_SIZE

    
def handle_changed():
    latest_changeddate_subquery = SlurpitImportedDevice.objects.filter(
        slurpit_id=OuterRef('slurpit_id')
    ).order_by('-changeddate').values('changeddate')[:1]
    qs = SlurpitStagedDevice.objects.annotate(
        latest_changeddate=Subquery(latest_changeddate_subquery)
    ).filter(
        changeddate__gt=F('latest_changeddate')
    )
    offset = 0
    count = len(qs)

    while offset < count:
        batch_qs = qs[offset:offset + BATCH_SIZE]
        to_import = []        
        for device in batch_qs:
            result = SlurpitImportedDevice.objects.get(slurpit_id=device.slurpit_id)
            result.copy_staged_values(device)
            
            result.save()
            get_create_dcim_objects(device)
            if result.mapped_device:
                if device.disabled == True:
                    if result.mapped_device.status != status_offline():
                        result.mapped_device.status=status_offline()
                # else:
                #     if result.mapped_device.status==status_offline():
                #         result.mapped_device.status=status_active()

                if device.serial:
                    result.mapped_device.serial = device.serial
                set_device_custom_fields(result.mapped_device, {
                    'slurpit_hostname': device.hostname,
                    'slurpit_fqdn': device.fqdn,
                    'slurpit_ipv4': device.ipv4,
                    'slurpit_serial': device.serial,
                    'slurpit_os_version': device.os_version,
                    'slurpit_snmp_uptime': device.snmp_uptime
                })   
                
                if device.ipv4:
                    prefix = Prefix.objects.filter(prefix__net_contains=device.ipv4).order_by('-prefix').first()
                    if prefix:
                        
                        address = f'{device.ipv4}/{prefix.prefix.prefixlen}'
                    else:
                        address = f'{device.ipv4}/32'
                    #### Remove Primary IPv4 on other device
                    other_device = Device.objects.filter(primary_ip4__address=address).first()
                    if other_device:
                        other_device.primary_ip4 = None
                        other_device.save()

                    interface = Interface.objects.filter(device=result.mapped_device)
                    if interface:
                        interface = interface.first()
                    else:
                        interface = Interface.objects.create(name='management1', device=result.mapped_device, type='other')

                    ipaddress = IPAddress.objects.filter(address=address)
                    if ipaddress:
                        ipaddress = ipaddress.first()
                    else:
                        ipaddress = IPAddress.objects.create(address=address, status='active')
                    
                    ipaddress.assigned_object = interface
                    ipaddress.save()
                    result.mapped_device.primary_ip4 = ipaddress

                result.mapped_device.name = device.hostname

                result.mapped_device.save()
        offset += BATCH_SIZE

def import_from_queryset(qs: QuerySet, **extra):
    count = len(qs)
    offset = 0

    while offset < count:
        batch_qs = qs[offset:offset + BATCH_SIZE]
        to_import = []        
        for device in batch_qs:
            device.mapped_device = get_dcim_device(device, **extra)
            to_import.append(device)
        SlurpitImportedDevice.objects.bulk_update(to_import, fields={'mapped_device_id'})
        offset += BATCH_SIZE

def get_dcim_device(staged: SlurpitStagedDevice | SlurpitImportedDevice, **extra) -> Device:
    kw = get_default_objects()
    cf = extra.pop(custom_field_data_name, {})
    interface_name = extra.pop('interface_name', 'management1')

    cf.update({
        'slurpit_hostname': staged.hostname,
        'slurpit_fqdn': staged.fqdn,
        'slurpit_platform': staged.device_os,
        'slurpit_manufacturer': staged.brand,
        'slurpit_devicetype': staged.device_type,
        'slurpit_ipv4': staged.ipv4,
        'slurpit_site': staged.site,
        'slurpit_serial': staged.serial,
        'slurpit_os_version': staged.os_version,
        'slurpit_snmp_uptime': staged.snmp_uptime
    })    

    try:
        platform = Platform.objects.get(name=staged.device_os)
    except:
        platform = Platform.objects.get(slug=slugify(staged.device_os))
    
    devicetype = None
    if 'device_type' in extra:
        devicetype = extra['device_type']
        staged.mapped_devicetype = extra['device_type']
        staged.save()

    elif 'device_type' not in extra and staged.mapped_devicetype is not None:
        devicetype = staged.mapped_devicetype
    
    if devicetype:
        platform = devicetype.default_platform
        
    kw.update({
        'name': staged.hostname,
        'platform': platform,
        custom_field_data_name: cf,
        **extra,
        # 'primary_ip4_id': int(ip_address(staged.fqdn)),
    })
    if staged.serial:
        kw['serial'] = staged.serial
    if 'device_type' not in extra and staged.mapped_devicetype is not None:
        kw['device_type'] = staged.mapped_devicetype
        
    if staged.disabled == True:
        kw.setdefault('status', status_offline())
    # else:
    #     kw.setdefault('status', status_active())
    
    device = Device.objects.filter(name=staged.hostname)

    if device:
        device = device.first()
        device.update(**kw)
    else:
        device = Device.objects.create(**kw)
    ensure_slurpit_tags(device)

    #Interface for new device.
    if staged.ipv4:
        prefix = Prefix.objects.filter(prefix__net_contains=staged.ipv4).order_by('-prefix').first()
        if prefix:
            address = f'{staged.ipv4}/{prefix.prefix.prefixlen}'
        else:
            address = f'{staged.ipv4}/32'
        #### Remove Primary IPv4 on other device
        other_device = Device.objects.filter(primary_ip4__address=address).first()
        if other_device:
            other_device.primary_ip4 = None
            other_device.save()
            
        interface, _ = Interface.objects.get_or_create(name=interface_name, device=device, defaults={'type':'other'})
        
        ipaddress = IPAddress.objects.filter(address=address)
        if ipaddress:
            ipaddress = ipaddress.first()
        else:
            ipaddress = IPAddress.objects.create(address=address, status='active')
        ipaddress.assigned_object = interface
        ipaddress.save()
        device.primary_ip4 = ipaddress

        device.save()
    
    return device

def get_from_staged(
        staged: SlurpitStagedDevice,
        add_dcim: bool
) -> SlurpitImportedDevice:
    device = SlurpitImportedDevice()
    device.copy_staged_values(staged)

    device.mapped_devicetype = get_create_dcim_objects(staged)
    if add_dcim:
        extra = {'device_type': device.mapped_devicetype} if device.mapped_devicetype else {}
        device.mapped_device = get_dcim_device(staged, **extra)
    return device


    
def get_latest_data_on_planning(hostname, planning_id):
    try:
        setting = SlurpitSetting.objects.get()
        uri_base = setting.server_url
        headers = {
                        'authorization': setting.api_key,
                        'useragent': f"{plugin_type}/requests",
                        'accept': 'application/json'
                    }
        uri_devices = f"{uri_base}/api/devices/snapshot/single/{hostname}/{planning_id}"

        r = requests.get(uri_devices, headers=headers, timeout=15, verify=False)
        # r.raise_for_status()
        if r.status_code != 200:
            return None

        data = r.json()
        log_message = f"Get the latest data from Slurp'it in {plugin_type.capitalize()} on planning ID."
        return data
    except ObjectDoesNotExist:
        setting = None
        log_message = "Need to set the setting parameter"
        return None

def import_plannings(plannings, delete=True):
    ids = {str(row['id']) : row for row in plannings if row['disabled'] == '0'}

    with transaction.atomic():
        if delete:
            count = SlurpitPlanning.objects.exclude(planning_id__in=ids.keys()).delete()[0]
            SlurpitSnapshot.objects.filter(planning_id__in=ids.keys()).delete()
            
        update_objects = SlurpitPlanning.objects.filter(planning_id__in=ids.keys())
        for planning in update_objects:
            obj = ids.pop(str(planning.planning_id))
            planning.name = obj['name']
            planning.comments = obj['comment']
            planning.save()
        
        to_save = []
        for obj in ids.values():
            to_save.append(SlurpitPlanning(name=obj['name'], comments=obj['comment'], planning_id=obj['id']))
        SlurpitPlanning.objects.bulk_create(to_save)
        


# Generated by Django 5.0.9 on 2024-10-30 09:10

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("slurpit_netbox", "0021_slurpitimporteddevice_site_slurpitstageddevice_site"),
    ]

    operations = [
        migrations.DeleteModel(
            name="SlurpitLog",
        ),
    ]



import django.db.models.deletion
from django.db import migrations, models


def copy_site_assignments(apps, schema_editor):
    """
    Copy site ForeignKey values to the scope GFK.
    """
    ContentType = apps.get_model('contenttypes', 'ContentType')
    SlurpitPrefix = apps.get_model('slurpit_netbox', 'SlurpitPrefix')
    Site = apps.get_model('dcim', 'Site')

    SlurpitPrefix.objects.filter(site__isnull=False).update(
        scope_type=ContentType.objects.get_for_model(Site), scope_id=models.F('site_id')
    )


class Migration(migrations.Migration):
    dependencies = [
        ('slurpit_netbox', '0023_alter_slurpitinterface_cable_end'),
    ]

    operations = [
        # Add the `scope` GenericForeignKey
        migrations.AddField(
            model_name='slurpitprefix',
            name='scope_id',
            field=models.PositiveBigIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='slurpitprefix',
            name='scope_type',
            field=models.ForeignKey(
                blank=True,
                limit_choices_to=models.Q(('model__in', ('region', 'sitegroup', 'site', 'location'))),
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='+',
                to='contenttypes.contenttype',
            ),
        ),
        # Copy over existing site assignments
        migrations.RunPython(code=copy_site_assignments, reverse_code=migrations.RunPython.noop),
    ]



from netbox.search import SearchIndex, register_search
from .models import SlurpitImportedDevice


@register_search
class SlurpitImportedDeviceIndex(SearchIndex):
    model = SlurpitImportedDevice
    fields = (
        ('hostname', 100),
        ('fqdn', 500),
    )



# Generated by Django 4.2.9 on 2024-05-01 02:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slurpit_netbox', '0009_slurpitinterface'),
    ]

    operations = [
        migrations.AlterField(
            model_name='slurpitinterface',
            name='name',
            field=models.CharField(blank=True, max_length=64),
        ),
    ]



# Generated by Django 5.1.5 on 2025-01-20 11:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slurpit_netbox', '0022_delete_slurpitlog'),
    ]

    operations = [
        migrations.AlterField(
            model_name='slurpitinterface',
            name='cable_end',
            field=models.CharField(blank=True, max_length=1, null=True),
        ),
    ]



# Generated by Django 4.2.9 on 2024-04-08 13:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('slurpit_netbox', '0005_manufacturer'),
    ]

    operations = [
        migrations.AddField(
            model_name='slurpitsnapshot',
            name='result_type',
            field=models.CharField(default='template_result', max_length=255),
        ),
    ]



# Generated by Django 4.2.9 on 2024-05-07 06:16

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('dcim', '0185_gfk_indexes'),
        ('slurpit_netbox', '0010_alter_slurpitinterface_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='slurpitinterface',
            name='device',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='%(class)ss', to='dcim.device'),
        ),
    ]
